from itertools import islice


# https://docs.python.org/3/library/itertools.html#itertools.batched
def batched(iterable, n):
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch



import asyncio
import importlib
import json
import logging
import re
import sys
from typing import List
import docker
import concurrent.futures

from jinja2 import Template


class DockerfileTemplate:
    """
    Manages the rendering of Dockerfile templates using Jinja2.

    Methods:
        configure: Prepares the template with specific configuration parameters.
        render: Renders the Dockerfile template with the provided configuration.
    """

    def __init__(self):
        self.template = self._load_template()
        self.configuration_params = {}

    def configure(
        self,
        extractor_path: "ExtractorPathWrapper",
        system_dependencies: List[str],
        python_dependencies: List[str],
        additional_pip_flags: str = "",
        dev: bool = False,
    ) -> "DockerfileTemplate":
        self.configuration_params = {
            "extractor_path": extractor_path.format(),
            "module_name": extractor_path.module_name,
            "module_file_name": extractor_path.file_name(),
            "class_name": extractor_path.class_name,
            "system_dependencies": system_dependencies,
            "python_dependencies": python_dependencies,
            "additional_pip_flags": additional_pip_flags,
            "dev": dev,
        }
        return self

    def _load_template(self) -> Template:
        from importlib import resources as impresources
        from . import dockerfiles

        inp_file = impresources.files(dockerfiles) / "Dockerfile.extractor"
        try:
            with open(inp_file) as f:
                template_content = f.read()
        except FileNotFoundError as e:
            raise FileNotFoundError(f"Failed to load template: {e}")

        return Template(template_content)

    def render(self, **kwargs) -> str:
        return self.template.render(**self.configuration_params, **kwargs)


class ExtractorPathWrapper:
    """
    Wraps the extractor's module and class names, providing formatting utilities.

    Attributes:
        module_name (str): The name of the module.
        class_name (str): The name of the class.

    Methods:
        format: Returns a formatted string combining module and class names.
        file_name: Returns the Python file name for the module.
    """

    def __init__(self, module_name: str, class_name: str):
        # i.e. "colbertv2"
        self.module_name = module_name
        # i.e. "ColBERTv2Base"
        self.class_name = class_name

    def validate(self) -> "ExtractorPathWrapper":
        # nothing to do for now
        return self

    def format(self) -> str:
        return f"{self.module_name_local()}:{self.class_name}"

    def module_name_local(self) -> str:
        # only the file name portion of the module name, without path.to.module
        file_name = self.module_name.replace("-", "_")
        return file_name.split(".")[-1]

    def file_name(self) -> str:
        return self.module_name_local() + ".py"


class DynamicModuleLoader:
    @staticmethod
    def load_module_from_path(module_name: str, file_path: str):
        """
        Dynamically loads a module from a given file path.

        Parameters:
            module_name (str): The name to assign to the loaded module.
            file_path (str): The file path of the Python module to load.

        Returns:
            The loaded module object.
        """
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
        return module


async def async_docker_build(client, **kwargs):
    """
    Asynchronously build a Docker image and stream build logs.

    Args:
        client: Docker client instance.
        **kwargs: Keyword arguments for the Docker build command.

    Returns:
        The Image object for the image that was built and a list of the build logs.

    Raises:
        docker.errors.BuildError if there is an error during the build.
    """
    loop = asyncio.get_event_loop()
    with concurrent.futures.ThreadPoolExecutor() as pool:
        # Run the Docker build process in a separate thread to avoid blocking
        resp = await loop.run_in_executor(pool, lambda: client.api.build(**kwargs))
        image_id = None
        build_logs = []

        streaming_log_handler = logging.StreamHandler(sys.stdout)
        streaming_log_handler.setLevel(logging.DEBUG)

        streaming_err_handler = logging.StreamHandler(sys.stderr)
        streaming_err_handler.setLevel(logging.ERROR)

        async for chunk in async_stream_logs(resp):
            build_logs.append(chunk)
            if "error" in chunk:
                streaming_err_handler.emit(
                    logging.LogRecord(
                        "docker",
                        logging.ERROR,
                        "docker",
                        0,
                        chunk["error"].strip(),
                        None,
                        None,
                    )
                )
                raise docker.errors.BuildError(chunk["error"], build_logs)
            if "stream" in chunk:
                streaming_log_handler.emit(
                    logging.LogRecord(
                        "docker",
                        logging.DEBUG,
                        "docker",
                        0,
                        chunk["stream"].strip(),
                        None,
                        None,
                    )
                )
                match = re.search(
                    r"(^Successfully built |sha256:)([0-9a-f]+)$", chunk["stream"]
                )
                if match:
                    image_id = match.group(2)

        if image_id:
            image = await loop.run_in_executor(
                pool, lambda: client.images.get(image_id)
            )
            return (image, build_logs)
        else:
            raise docker.errors.BuildError("Unknown error", build_logs)


async def async_stream_logs(resp):
    """
    Asynchronously stream Docker build logs.

    Args:
        resp: The response object from the Docker API build request.

    Yields:
        Decoded JSON objects from the build logs.
    """
    for line in resp:
        # a line can have multiple json objects, connected with \r\n
        for line in line.decode("utf-8").split("\r\n"):
            if line:
                yield json.loads(line)



from abc import ABC, abstractmethod
from typing import Tuple, Dict, List, Type, Optional
import json
from importlib import import_module
from typing import get_type_hints, Literal, Union, Dict

from pydantic import BaseModel, Json
from genson import SchemaBuilder
import requests
import os


class EmbeddingSchema(BaseModel):
    dim: int
    distance: str


class Embedding(BaseModel):
    values: List[float]
    distance: str


class ExtractorDescription(BaseModel):
    name: str
    version: str
    description: str
    python_dependencies: List[str]
    system_dependencies: List[str]
    embedding_schemas: dict[str, EmbeddingSchema]
    metadata_schemas: dict[str, Dict]
    input_params: Optional[str]
    input_mime_types: List[str]


class Feature(BaseModel):
    feature_type: Literal["embedding", "metadata"]
    name: str
    value: Json

    @classmethod
    def embedding(cls, values: List[float], name: str = "embedding", distance="cosine"):
        embedding = Embedding(values=values, distance=distance)
        return cls(
            feature_type="embedding",
            name=name,
            value=embedding.model_dump_json(),
        )

    @classmethod
    def metadata(cls, value: Json, name: str = "metadata"):
        return cls(feature_type="metadata", name=name, value=json.dumps(value))

    def get_schema(self) -> dict:
        if self.feature_type == "embedding":
            embedding_value: Embedding = Embedding.model_validate_json(self.value)
            return EmbeddingSchema(
                dim=len(embedding_value.values), distance=embedding_value.distance
            ).model_dump()
        else:
            schema_builder = SchemaBuilder()
            schema_builder.add_object(self.value)
            metadata_schema = schema_builder.to_schema()
            schema = {}
            print(metadata_schema)
            for k, v in metadata_schema["properties"].items():
                schema[k] = {"type": v["type"]}
            return schema


class Content(BaseModel):
    content_type: Optional[str]
    data: bytes
    features: List[Feature] = []
    labels: Dict[str, str] = {}

    @classmethod
    def from_text(
        cls, text: str, features: List[Feature] = [], labels: Dict[str, str] = {}
    ):
        return cls(
            content_type="text/plain",
            data=bytes(text, "utf-8"),
            features=features,
            labels=labels,
        )

    @classmethod
    def from_file(cls, path: str):
        import mimetypes

        m, _ = mimetypes.guess_type(path)
        with open(path, "rb") as f:
            return cls(content_type=m, data=f.read())


class Extractor(ABC):
    name: str = ""

    version: str = "0.0.0"

    system_dependencies: List[str] = []

    python_dependencies: List[str] = []

    description: str = ""

    input_mime_types = ["text/plain"]

    @abstractmethod
    def extract(
        self, content: Content, params: Type[BaseModel] = None
    ) -> List[Union[Feature, Content]]:
        """
        Extracts information from the content. Returns a list of features to add
        to the content.
        It can also return a list of Content objects, which will be added to storage
        and any extraction policies defined will be applied to them.
        """
        pass

    @abstractmethod
    def sample_input(self) -> Tuple[Content, Type[BaseModel]]:
        pass

    def extract_sample_input(self) -> List[Union[Feature, Content]]:
        input = self.sample_input()
        return self.extract(*input)

    def _download_file(self, url, filename):
        if os.path.exists(filename):
            # file exists skip
            return
        try:
            with requests.get(url, stream=True) as r:
                r.raise_for_status()  # Raises an HTTPError if the response status code is 4XX/5XX
                with open(filename, "wb") as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)
        except requests.exceptions.RequestException as e:
            print(f"Error downloading the file: {e}")

    def sample_mp3(self, features: List[Feature] = []) -> Content:
        file_name = "sample.mp3"
        self._download_file(
            "https://extractor-files.diptanu-6d5.workers.dev/sample-000009.mp3",
            file_name,
        )
        f = open(file_name, "rb")
        return Content(content_type="audio/mpeg", data=f.read(), features=features)

    def sample_mp4(self, features: List[Feature] = []) -> Content:
        file_name = "sample.mp4"
        self._download_file(
            "https://extractor-files.diptanu-6d5.workers.dev/sample.mp4",
            file_name,
        )
        f = open(file_name, "rb")
        return Content(content_type="video/mp4", data=f.read(), features=features)

    def sample_jpg(self, features: List[Feature] = []) -> Content:
        file_name = "sample.jpg"
        self._download_file(
            "https://extractor-files.diptanu-6d5.workers.dev/people-standing.jpg",
            file_name,
        )
        f = open(file_name, "rb")
        return Content(content_type="image/jpg", data=f.read(), features=features)

    def sample_invoice_jpg(self, features: List[Feature] = []) -> Content:
        file_name = "sample.jpg"
        self._download_file(
            "https://extractor-files.diptanu-6d5.workers.dev/invoice-example.jpg",
            file_name,
        )
        f = open(file_name, "rb")
        return Content(content_type="image/jpg", data=f.read(), features=features)

    def sample_invoice_pdf(self, features: List[Feature] = []) -> Content:
        file_name = "sample.pdf"
        self._download_file(
            "https://extractor-files.diptanu-6d5.workers.dev/invoice-example.pdf",
            file_name,
        )
        f = open(file_name, "rb")
        return Content(content_type="application/pdf", data=f.read(), features=features)

    def sample_image_based_pdf(self, features: List[Feature] = []) -> Content:
        file_name = "sample.pdf"
        self._download_file(
            "https://extractor-files.diptanu-6d5.workers.dev/image-based.pdf",
            file_name,
        )
        f = open(file_name, "rb")
        return Content(content_type="application/pdf", data=f.read(), features=features)

    def sample_scientific_pdf(self, features: List[Feature] = []) -> Content:
        file_name = "sample.pdf"
        self._download_file(
            "https://extractor-files.diptanu-6d5.workers.dev/scientific-paper-example.pdf",
            file_name,
        )
        f = open(file_name, "rb")
        return Content(content_type="application/pdf", data=f.read(), features=features)


class ExtractorWrapper:
    def __init__(self, module_name: str, class_name: str):
        module = import_module(module_name)
        cls = getattr(module, class_name)
        self._instance: Extractor = cls()
        self._param_cls = get_type_hints(cls.extract).get("params", None)
        extract_batch = getattr(self._instance, "extract_batch", None)
        self._has_batch_extract = True if callable(extract_batch) else False

    def extract_batch(
        self, content_list: Dict[str, Content], params: Json
    ) -> Dict[str, List[Union[Feature, Content]]]:
        params_dict = json.loads(params) if params is not None else {}
        param_instance = (
            self._param_cls.model_validate(params_dict)
            if self._param_cls is not None
            else None
        )
        if self._has_batch_extract:
            keys = []
            values = []
            for k, v in content_list.items():
                keys.append(k)
                values.append(v)
            result = self._instance.extract_batch(values, param_instance)
            out: Dict[str, List[Union[Feature, Content]]] = {}
            for i, extractor_out in enumerate(result):
                out[keys[i]] = extractor_out
            return out
        out = {}
        for task_id, content in content_list.items():
            out[task_id] = self._instance.extract(content, param_instance)
        return out

    def describe(self, input_params: Type[BaseModel] = None) -> ExtractorDescription:
        s_input = self._instance.sample_input()
        if type(s_input) == tuple:
            (s_input, input_params) = s_input
        # Come back to this when we can support schemas based on user defined input params
        if input_params is None:
            input_params = (
                self._param_cls().model_dump_json()
                if self._param_cls is not None
                else None
            )
        outputs: Dict[str, List[Union[Feature, Content]]] = self.extract_batch(
            {"task_id": s_input}, input_params
        )
        embedding_schemas = {}
        metadata_schemas = {}
        json_schema = (
            self._param_cls.model_json_schema() if self._param_cls is not None else None
        )
        output = outputs["task_id"]
        for out in output:
            features = out.features if type(out) == Content else [out]
            for feature in features:
                if feature.feature_type == "embedding":
                    embedding_value: Embedding = Embedding.model_validate(feature.value)
                    embedding_schema = EmbeddingSchema(
                        dim=len(embedding_value.values),
                        distance=embedding_value.distance,
                    )
                    embedding_schemas[feature.name] = embedding_schema
                elif feature.feature_type == "metadata":
                    metadata_schemas[feature.name] = feature.get_schema()
        return ExtractorDescription(
            name=self._instance.name,
            version=self._instance.version,
            description=self._instance.description,
            python_dependencies=self._instance.python_dependencies,
            system_dependencies=self._instance.system_dependencies,
            embedding_schemas=embedding_schemas,
            metadata_schemas=metadata_schemas,
            input_mime_types=self._instance.input_mime_types,
            input_params=json.dumps(json_schema),
        )



import boto3

from .base_extractor import Content
from . import coordinator_service_pb2
from urllib.parse import urlparse
from azure.storage.blob import BlobServiceClient
from azure.identity import DefaultAzureCredential
from google.cloud import storage
import httpx
from typing import Dict
import asyncio


def disk_loader(file_path: str):
    print(file_path)
    file_path = file_path.removeprefix("file:/")
    with open(file_path, "rb") as f:
        return f.read()


def s3_loader(s3_url: str) -> bytes:
    parsed_url = urlparse(s3_url)
    bucket_name = parsed_url.netloc
    key = parsed_url.path.lstrip("/")

    s3 = boto3.client("s3")

    response = s3.get_object(Bucket=bucket_name, Key=key)
    return response["Body"].read()


def azure_blob_loader(blob_url: str) -> bytes:
    token_credential = DefaultAzureCredential()
    parsed_url = urlparse(blob_url)
    account_url = f"https://{parsed_url.netloc}"
    container_name = parsed_url.path.split("/")[1]
    blob_name = "/".join(parsed_url.path.split("/")[2:])

    blob_service_client = BlobServiceClient(
        account_url=account_url, credential=token_credential
    )
    blob_client = blob_service_client.get_blob_client(
        container=container_name, blob=blob_name
    )

    return blob_client.download_blob().readall()


def gcp_storage_loader(storage_url: str) -> bytes:
    parsed_url = urlparse(storage_url)
    bucket_name = parsed_url.netloc
    blob_name = parsed_url.path.lstrip("/")

    client = storage.Client()
    bucket = client.get_bucket(bucket_name)

    blob = bucket.blob(blob_name)
    return blob.download_as_bytes()


async def fetch_url(id, url):
    try:
        async with httpx.AsyncClient() as client:
            print(f"downloading url {url}")
            response = await client.get(url, follow_redirects=True)
            response.raise_for_status()
            return id, response.read()
    except Exception as e:
        return id, e


async def download_parallel(urls):
    tasks = [fetch_url(id, url) for id, url in urls.items()]
    results = await asyncio.gather(*tasks, return_exceptions=True)
    return results


async def download_content(urls: Dict[str, str]) -> Dict[str, Content]:
    out = {}
    disk_urls = {}
    s3_urls = {}
    http_urls = {}
    gs_urls = {}
    for task_id, url in urls.items():
        if url.startswith("file://"):
            disk_urls[task_id] = url
        elif url.startswith("s3://"):
            s3_urls[task_id] = url
        elif url.startswith("https://") or url.startswith("http://"):
            http_urls[task_id] = url
        elif url.startswith("gs://"):
            gs_urls[task_id] = url
        else:
            out[task_id] = Exception(f"unsupported storage url {url}")
    for task_id, url in gs_urls.items():
        out[task_id] = gcp_storage_loader(url)
    for task_id, url in s3_urls.items():
        out[task_id] = s3_loader(url)
    for task_id, url in disk_urls.items():
        out[task_id] = disk_loader(url)

    result = await download_parallel(http_urls)
    for task_id, b in result:
        out[task_id] = b

    return out


def create_content(bytes, task: coordinator_service_pb2.Task) -> Content:
    metadata = task.content_metadata
    return Content(
        content_type=metadata.mime, data=bytes, features=[], labels=metadata.labels
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: coordinator_service.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x19\x63oordinator_service.proto\x12\x14indexify_coordinator"1\n\x19GetContentMetadataRequest\x12\x14\n\x0c\x63ontent_list\x18\x01 \x03(\t"Y\n\x1aGetContentMetadataResponse\x12;\n\x0c\x63ontent_list\x18\x01 \x03(\x0b\x32%.indexify_coordinator.ContentMetadata"\xaa\x01\n\x11UpdateTaskRequest\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t\x12\x0f\n\x07task_id\x18\x02 \x01(\t\x12\x32\n\x07outcome\x18\x03 \x01(\x0e\x32!.indexify_coordinator.TaskOutcome\x12;\n\x0c\x63ontent_list\x18\x04 \x03(\x0b\x32%.indexify_coordinator.ContentMetadata"\x19\n\x17ListStateChangesRequest"k\n\x0bStateChange\x12\n\n\x02id\x18\x01 \x01(\t\x12\x11\n\tobject_id\x18\x02 \x01(\t\x12\x13\n\x0b\x63hange_type\x18\x03 \x01(\t\x12\x12\n\ncreated_at\x18\x04 \x01(\x04\x12\x14\n\x0cprocessed_at\x18\x05 \x01(\x04"N\n\x18ListStateChangesResponse\x12\x32\n\x07\x63hanges\x18\x01 \x03(\x0b\x32!.indexify_coordinator.StateChange"@\n\x10ListTasksRequest\x12\x11\n\tnamespace\x18\x01 \x01(\t\x12\x19\n\x11\x65xtraction_policy\x18\x02 \x01(\t">\n\x11ListTasksResponse\x12)\n\x05tasks\x18\x01 \x03(\x0b\x32\x1a.indexify_coordinator.Task"\x14\n\x12UpdateTaskResponse"3\n\x1eGetExtractorCoordinatesRequest\x12\x11\n\textractor\x18\x02 \x01(\t"0\n\x1fGetExtractorCoordinatesResponse\x12\r\n\x05\x61\x64\x64rs\x18\x01 \x03(\t"\'\n\x12ListIndexesRequest\x12\x11\n\tnamespace\x18\x01 \x01(\t"C\n\x13ListIndexesResponse\x12,\n\x07indexes\x18\x01 \x03(\x0b\x32\x1b.indexify_coordinator.Index"2\n\x0fGetIndexRequest\x12\x11\n\tnamespace\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t">\n\x10GetIndexResponse\x12*\n\x05index\x18\x01 \x01(\x0b\x32\x1b.indexify_coordinator.Index"@\n\x12\x43reateIndexRequest\x12*\n\x05index\x18\x02 \x01(\x0b\x32\x1b.indexify_coordinator.Index"\x15\n\x13\x43reateIndexResponse"z\n\x05Index\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tnamespace\x18\x02 \x01(\t\x12\x12\n\ntable_name\x18\x03 \x01(\t\x12\x0e\n\x06schema\x18\x04 \x01(\t\x12\x19\n\x11\x65xtraction_policy\x18\x05 \x01(\t\x12\x11\n\textractor\x18\x06 \x01(\t"\x1e\n\tEmbedding\x12\x11\n\tembedding\x18\x01 \x03(\x02" \n\nAttributes\x12\x12\n\nattributes\x18\x02 \x01(\t"\x81\x01\n\x07\x46\x65\x61ture\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x32\n\tembedding\x18\x02 \x01(\x0b\x32\x1f.indexify_coordinator.Embedding\x12\x34\n\nattributes\x18\x03 \x01(\x0b\x32 .indexify_coordinator.Attributes"V\n\x07\x43ontent\x12\x0c\n\x04mime\x18\x02 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x12/\n\x08\x66\x65\x61tures\x18\x04 \x03(\x0b\x32\x1d.indexify_coordinator.Feature"p\n\x17RegisterExecutorRequest\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t\x12\x0c\n\x04\x61\x64\x64r\x18\x02 \x01(\t\x12\x32\n\textractor\x18\x03 \x01(\x0b\x32\x1f.indexify_coordinator.Extractor"/\n\x18RegisterExecutorResponse\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t">\n\x10HeartbeatRequest\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t\x12\x15\n\rpending_tasks\x18\x02 \x01(\x03"S\n\x11HeartbeatResponse\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t\x12)\n\x05tasks\x18\x02 \x03(\x0b\x32\x1a.indexify_coordinator.Task"\xeb\x02\n\x04Task\x12\n\n\x02id\x18\x01 \x01(\t\x12\x11\n\textractor\x18\x02 \x01(\t\x12\x11\n\tnamespace\x18\x03 \x01(\t\x12?\n\x10\x63ontent_metadata\x18\x04 \x01(\x0b\x32%.indexify_coordinator.ContentMetadata\x12\x14\n\x0cinput_params\x18\x05 \x01(\t\x12\x19\n\x11\x65xtraction_policy\x18\x06 \x01(\t\x12P\n\x14output_index_mapping\x18\x07 \x03(\x0b\x32\x32.indexify_coordinator.Task.OutputIndexMappingEntry\x12\x32\n\x07outcome\x18\x08 \x01(\x0e\x32!.indexify_coordinator.TaskOutcome\x1a\x39\n\x17OutputIndexMappingEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\x17\n\x15ListExtractorsRequest"M\n\x16ListExtractorsResponse\x12\x33\n\nextractors\x18\x01 \x03(\x0b\x32\x1f.indexify_coordinator.Extractor"\xf1\x02\n\tExtractor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x14\n\x0cinput_params\x18\x03 \x01(\t\x12P\n\x11\x65mbedding_schemas\x18\x04 \x03(\x0b\x32\x35.indexify_coordinator.Extractor.EmbeddingSchemasEntry\x12N\n\x10metadata_schemas\x18\x05 \x03(\x0b\x32\x34.indexify_coordinator.Extractor.MetadataSchemasEntry\x12\x18\n\x10input_mime_types\x18\x06 \x03(\t\x1a\x37\n\x15\x45mbeddingSchemasEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x36\n\x14MetadataSchemasEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"#\n\x13GetNamespaceRequest\x12\x0c\n\x04name\x18\x01 \x01(\t"J\n\x14GetNamespaceResponse\x12\x32\n\tnamespace\x18\x01 \x01(\x0b\x32\x1f.indexify_coordinator.Namespace"\xc6\x01\n\x12ListContentRequest\x12\x11\n\tnamespace\x18\x01 \x01(\t\x12\x0e\n\x06source\x18\x02 \x01(\t\x12\x11\n\tparent_id\x18\x03 \x01(\t\x12I\n\tlabels_eq\x18\x04 \x03(\x0b\x32\x36.indexify_coordinator.ListContentRequest.LabelsEqEntry\x1a/\n\rLabelsEqEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"R\n\x13ListContentResponse\x12;\n\x0c\x63ontent_list\x18\x01 \x03(\x0b\x32%.indexify_coordinator.ContentMetadata"2\n\x1dListExtractionPoliciesRequest\x12\x11\n\tnamespace\x18\x01 \x01(\t"Z\n\x1eListExtractionPoliciesResponse\x12\x38\n\x08policies\x18\x01 \x03(\x0b\x32&.indexify_coordinator.ExtractionPolicy"`\n\x16\x43reateNamespaceRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x38\n\x08policies\x18\x02 \x03(\x0b\x32&.indexify_coordinator.ExtractionPolicy";\n\x17\x43reateNamespaceResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x12\n\ncreated_at\x18\x02 \x01(\x03"\x16\n\x14ListNamespaceRequest"L\n\x15ListNamespaceResponse\x12\x33\n\nnamespaces\x18\x01 \x03(\x0b\x32\x1f.indexify_coordinator.Namespace"\xd7\x01\n\x10\x45xtractionPolicy\x12\x11\n\textractor\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x14\n\x0cinput_params\x18\x04 \x01(\t\x12\x44\n\x07\x66ilters\x18\x05 \x03(\x0b\x32\x33.indexify_coordinator.ExtractionPolicy.FiltersEntry\x12\x16\n\x0e\x63ontent_source\x18\x06 \x01(\t\x1a.\n\x0c\x46iltersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"x\n\x17\x45xtractionPolicyRequest\x12\x11\n\tnamespace\x18\x01 \x01(\t\x12\x36\n\x06policy\x18\x03 \x01(\x0b\x32&.indexify_coordinator.ExtractionPolicy\x12\x12\n\ncreated_at\x18\x02 \x01(\x03"\xbb\x03\n\x18\x45xtractionPolicyResponse\x12\x12\n\ncreated_at\x18\x03 \x01(\x03\x12\x32\n\textractor\x18\x02 \x01(\x0b\x32\x1f.indexify_coordinator.Extractor\x12k\n\x18index_name_table_mapping\x18\x04 \x03(\x0b\x32I.indexify_coordinator.ExtractionPolicyResponse.IndexNameTableMappingEntry\x12m\n\x19output_index_name_mapping\x18\x05 \x03(\x0b\x32J.indexify_coordinator.ExtractionPolicyResponse.OutputIndexNameMappingEntry\x1a<\n\x1aIndexNameTableMappingEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a=\n\x1bOutputIndexNameMappingEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\xa3\x02\n\x0f\x43ontentMetadata\x12\n\n\x02id\x18\x01 \x01(\t\x12\x11\n\tfile_name\x18\x02 \x01(\t\x12\x11\n\tparent_id\x18\x03 \x01(\t\x12\x0c\n\x04mime\x18\x04 \x01(\t\x12\x41\n\x06labels\x18\x05 \x03(\x0b\x32\x31.indexify_coordinator.ContentMetadata.LabelsEntry\x12\x13\n\x0bstorage_url\x18\x06 \x01(\t\x12\x12\n\ncreated_at\x18\x07 \x01(\x03\x12\x11\n\tnamespace\x18\x08 \x01(\t\x12\x0e\n\x06source\x18\t \x01(\t\x12\x12\n\nsize_bytes\x18\n \x01(\x04\x1a-\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"N\n\x14\x43reateContentRequest\x12\x36\n\x07\x63ontent\x18\x02 \x01(\x0b\x32%.indexify_coordinator.ContentMetadata"#\n\x15\x43reateContentResponse\x12\n\n\x02id\x18\x01 \x01(\t"S\n\tNamespace\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x38\n\x08policies\x18\x02 \x03(\x0b\x32&.indexify_coordinator.ExtractionPolicy"=\n\x10GetSchemaRequest\x12\x11\n\tnamespace\x18\x01 \x01(\t\x12\x16\n\x0e\x63ontent_source\x18\x02 \x01(\t"O\n\x11GetSchemaResponse\x12:\n\x06schema\x18\x01 \x01(\x0b\x32*.indexify_coordinator.StructuredDataSchema"?\n\x14StructuredDataSchema\x12\x0f\n\x07\x63olumns\x18\x01 \x01(\t\x12\x16\n\x0e\x63ontent_source\x18\x02 \x01(\t"(\n\x13GetAllSchemaRequest\x12\x11\n\tnamespace\x18\x01 \x01(\t"S\n\x14GetAllSchemaResponse\x12;\n\x07schemas\x18\x01 \x03(\x0b\x32*.indexify_coordinator.StructuredDataSchema*3\n\x0bTaskOutcome\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06\x46\x41ILED\x10\x01\x12\x0b\n\x07SUCCESS\x10\x02\x32\x92\x11\n\x12\x43oordinatorService\x12j\n\rCreateContent\x12*.indexify_coordinator.CreateContentRequest\x1a+.indexify_coordinator.CreateContentResponse"\x00\x12y\n\x12GetContentMetadata\x12/.indexify_coordinator.GetContentMetadataRequest\x1a\x30.indexify_coordinator.GetContentMetadataResponse"\x00\x12\x64\n\x0bListContent\x12(.indexify_coordinator.ListContentRequest\x1a).indexify_coordinator.ListContentResponse"\x00\x12y\n\x16\x43reateExtractionPolicy\x12-.indexify_coordinator.ExtractionPolicyRequest\x1a..indexify_coordinator.ExtractionPolicyResponse"\x00\x12\x85\x01\n\x16ListExtractionPolicies\x12\x33.indexify_coordinator.ListExtractionPoliciesRequest\x1a\x34.indexify_coordinator.ListExtractionPoliciesResponse"\x00\x12i\n\x08\x43reateNS\x12,.indexify_coordinator.CreateNamespaceRequest\x1a-.indexify_coordinator.CreateNamespaceResponse"\x00\x12\x63\n\x06ListNS\x12*.indexify_coordinator.ListNamespaceRequest\x1a+.indexify_coordinator.ListNamespaceResponse"\x00\x12`\n\x05GetNS\x12).indexify_coordinator.GetNamespaceRequest\x1a*.indexify_coordinator.GetNamespaceResponse"\x00\x12m\n\x0eListExtractors\x12+.indexify_coordinator.ListExtractorsRequest\x1a,.indexify_coordinator.ListExtractorsResponse"\x00\x12s\n\x10RegisterExecutor\x12-.indexify_coordinator.RegisterExecutorRequest\x1a..indexify_coordinator.RegisterExecutorResponse"\x00\x12\x62\n\tHeartbeat\x12&.indexify_coordinator.HeartbeatRequest\x1a\'.indexify_coordinator.HeartbeatResponse"\x00(\x01\x30\x01\x12\x64\n\x0bListIndexes\x12(.indexify_coordinator.ListIndexesRequest\x1a).indexify_coordinator.ListIndexesResponse"\x00\x12[\n\x08GetIndex\x12%.indexify_coordinator.GetIndexRequest\x1a&.indexify_coordinator.GetIndexResponse"\x00\x12\x64\n\x0b\x43reateIndex\x12(.indexify_coordinator.CreateIndexRequest\x1a).indexify_coordinator.CreateIndexResponse"\x00\x12\x88\x01\n\x17GetExtractorCoordinates\x12\x34.indexify_coordinator.GetExtractorCoordinatesRequest\x1a\x35.indexify_coordinator.GetExtractorCoordinatesResponse"\x00\x12\x61\n\nUpdateTask\x12\'.indexify_coordinator.UpdateTaskRequest\x1a(.indexify_coordinator.UpdateTaskResponse"\x00\x12s\n\x10ListStateChanges\x12-.indexify_coordinator.ListStateChangesRequest\x1a..indexify_coordinator.ListStateChangesResponse"\x00\x12^\n\tListTasks\x12&.indexify_coordinator.ListTasksRequest\x1a\'.indexify_coordinator.ListTasksResponse"\x00\x12^\n\tGetSchema\x12&.indexify_coordinator.GetSchemaRequest\x1a\'.indexify_coordinator.GetSchemaResponse"\x00\x12\x66\n\x0bListSchemas\x12).indexify_coordinator.GetAllSchemaRequest\x1a*.indexify_coordinator.GetAllSchemaResponse"\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "coordinator_service_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_TASK_OUTPUTINDEXMAPPINGENTRY"]._options = None
    _globals["_TASK_OUTPUTINDEXMAPPINGENTRY"]._serialized_options = b"8\001"
    _globals["_EXTRACTOR_EMBEDDINGSCHEMASENTRY"]._options = None
    _globals["_EXTRACTOR_EMBEDDINGSCHEMASENTRY"]._serialized_options = b"8\001"
    _globals["_EXTRACTOR_METADATASCHEMASENTRY"]._options = None
    _globals["_EXTRACTOR_METADATASCHEMASENTRY"]._serialized_options = b"8\001"
    _globals["_LISTCONTENTREQUEST_LABELSEQENTRY"]._options = None
    _globals["_LISTCONTENTREQUEST_LABELSEQENTRY"]._serialized_options = b"8\001"
    _globals["_EXTRACTIONPOLICY_FILTERSENTRY"]._options = None
    _globals["_EXTRACTIONPOLICY_FILTERSENTRY"]._serialized_options = b"8\001"
    _globals["_EXTRACTIONPOLICYRESPONSE_INDEXNAMETABLEMAPPINGENTRY"]._options = None
    _globals[
        "_EXTRACTIONPOLICYRESPONSE_INDEXNAMETABLEMAPPINGENTRY"
    ]._serialized_options = b"8\001"
    _globals["_EXTRACTIONPOLICYRESPONSE_OUTPUTINDEXNAMEMAPPINGENTRY"]._options = None
    _globals[
        "_EXTRACTIONPOLICYRESPONSE_OUTPUTINDEXNAMEMAPPINGENTRY"
    ]._serialized_options = b"8\001"
    _globals["_CONTENTMETADATA_LABELSENTRY"]._options = None
    _globals["_CONTENTMETADATA_LABELSENTRY"]._serialized_options = b"8\001"
    _globals["_TASKOUTCOME"]._serialized_start = 5137
    _globals["_TASKOUTCOME"]._serialized_end = 5188
    _globals["_GETCONTENTMETADATAREQUEST"]._serialized_start = 51
    _globals["_GETCONTENTMETADATAREQUEST"]._serialized_end = 100
    _globals["_GETCONTENTMETADATARESPONSE"]._serialized_start = 102
    _globals["_GETCONTENTMETADATARESPONSE"]._serialized_end = 191
    _globals["_UPDATETASKREQUEST"]._serialized_start = 194
    _globals["_UPDATETASKREQUEST"]._serialized_end = 364
    _globals["_LISTSTATECHANGESREQUEST"]._serialized_start = 366
    _globals["_LISTSTATECHANGESREQUEST"]._serialized_end = 391
    _globals["_STATECHANGE"]._serialized_start = 393
    _globals["_STATECHANGE"]._serialized_end = 500
    _globals["_LISTSTATECHANGESRESPONSE"]._serialized_start = 502
    _globals["_LISTSTATECHANGESRESPONSE"]._serialized_end = 580
    _globals["_LISTTASKSREQUEST"]._serialized_start = 582
    _globals["_LISTTASKSREQUEST"]._serialized_end = 646
    _globals["_LISTTASKSRESPONSE"]._serialized_start = 648
    _globals["_LISTTASKSRESPONSE"]._serialized_end = 710
    _globals["_UPDATETASKRESPONSE"]._serialized_start = 712
    _globals["_UPDATETASKRESPONSE"]._serialized_end = 732
    _globals["_GETEXTRACTORCOORDINATESREQUEST"]._serialized_start = 734
    _globals["_GETEXTRACTORCOORDINATESREQUEST"]._serialized_end = 785
    _globals["_GETEXTRACTORCOORDINATESRESPONSE"]._serialized_start = 787
    _globals["_GETEXTRACTORCOORDINATESRESPONSE"]._serialized_end = 835
    _globals["_LISTINDEXESREQUEST"]._serialized_start = 837
    _globals["_LISTINDEXESREQUEST"]._serialized_end = 876
    _globals["_LISTINDEXESRESPONSE"]._serialized_start = 878
    _globals["_LISTINDEXESRESPONSE"]._serialized_end = 945
    _globals["_GETINDEXREQUEST"]._serialized_start = 947
    _globals["_GETINDEXREQUEST"]._serialized_end = 997
    _globals["_GETINDEXRESPONSE"]._serialized_start = 999
    _globals["_GETINDEXRESPONSE"]._serialized_end = 1061
    _globals["_CREATEINDEXREQUEST"]._serialized_start = 1063
    _globals["_CREATEINDEXREQUEST"]._serialized_end = 1127
    _globals["_CREATEINDEXRESPONSE"]._serialized_start = 1129
    _globals["_CREATEINDEXRESPONSE"]._serialized_end = 1150
    _globals["_INDEX"]._serialized_start = 1152
    _globals["_INDEX"]._serialized_end = 1274
    _globals["_EMBEDDING"]._serialized_start = 1276
    _globals["_EMBEDDING"]._serialized_end = 1306
    _globals["_ATTRIBUTES"]._serialized_start = 1308
    _globals["_ATTRIBUTES"]._serialized_end = 1340
    _globals["_FEATURE"]._serialized_start = 1343
    _globals["_FEATURE"]._serialized_end = 1472
    _globals["_CONTENT"]._serialized_start = 1474
    _globals["_CONTENT"]._serialized_end = 1560
    _globals["_REGISTEREXECUTORREQUEST"]._serialized_start = 1562
    _globals["_REGISTEREXECUTORREQUEST"]._serialized_end = 1674
    _globals["_REGISTEREXECUTORRESPONSE"]._serialized_start = 1676
    _globals["_REGISTEREXECUTORRESPONSE"]._serialized_end = 1723
    _globals["_HEARTBEATREQUEST"]._serialized_start = 1725
    _globals["_HEARTBEATREQUEST"]._serialized_end = 1787
    _globals["_HEARTBEATRESPONSE"]._serialized_start = 1789
    _globals["_HEARTBEATRESPONSE"]._serialized_end = 1872
    _globals["_TASK"]._serialized_start = 1875
    _globals["_TASK"]._serialized_end = 2238
    _globals["_TASK_OUTPUTINDEXMAPPINGENTRY"]._serialized_start = 2181
    _globals["_TASK_OUTPUTINDEXMAPPINGENTRY"]._serialized_end = 2238
    _globals["_LISTEXTRACTORSREQUEST"]._serialized_start = 2240
    _globals["_LISTEXTRACTORSREQUEST"]._serialized_end = 2263
    _globals["_LISTEXTRACTORSRESPONSE"]._serialized_start = 2265
    _globals["_LISTEXTRACTORSRESPONSE"]._serialized_end = 2342
    _globals["_EXTRACTOR"]._serialized_start = 2345
    _globals["_EXTRACTOR"]._serialized_end = 2714
    _globals["_EXTRACTOR_EMBEDDINGSCHEMASENTRY"]._serialized_start = 2603
    _globals["_EXTRACTOR_EMBEDDINGSCHEMASENTRY"]._serialized_end = 2658
    _globals["_EXTRACTOR_METADATASCHEMASENTRY"]._serialized_start = 2660
    _globals["_EXTRACTOR_METADATASCHEMASENTRY"]._serialized_end = 2714
    _globals["_GETNAMESPACEREQUEST"]._serialized_start = 2716
    _globals["_GETNAMESPACEREQUEST"]._serialized_end = 2751
    _globals["_GETNAMESPACERESPONSE"]._serialized_start = 2753
    _globals["_GETNAMESPACERESPONSE"]._serialized_end = 2827
    _globals["_LISTCONTENTREQUEST"]._serialized_start = 2830
    _globals["_LISTCONTENTREQUEST"]._serialized_end = 3028
    _globals["_LISTCONTENTREQUEST_LABELSEQENTRY"]._serialized_start = 2981
    _globals["_LISTCONTENTREQUEST_LABELSEQENTRY"]._serialized_end = 3028
    _globals["_LISTCONTENTRESPONSE"]._serialized_start = 3030
    _globals["_LISTCONTENTRESPONSE"]._serialized_end = 3112
    _globals["_LISTEXTRACTIONPOLICIESREQUEST"]._serialized_start = 3114
    _globals["_LISTEXTRACTIONPOLICIESREQUEST"]._serialized_end = 3164
    _globals["_LISTEXTRACTIONPOLICIESRESPONSE"]._serialized_start = 3166
    _globals["_LISTEXTRACTIONPOLICIESRESPONSE"]._serialized_end = 3256
    _globals["_CREATENAMESPACEREQUEST"]._serialized_start = 3258
    _globals["_CREATENAMESPACEREQUEST"]._serialized_end = 3354
    _globals["_CREATENAMESPACERESPONSE"]._serialized_start = 3356
    _globals["_CREATENAMESPACERESPONSE"]._serialized_end = 3415
    _globals["_LISTNAMESPACEREQUEST"]._serialized_start = 3417
    _globals["_LISTNAMESPACEREQUEST"]._serialized_end = 3439
    _globals["_LISTNAMESPACERESPONSE"]._serialized_start = 3441
    _globals["_LISTNAMESPACERESPONSE"]._serialized_end = 3517
    _globals["_EXTRACTIONPOLICY"]._serialized_start = 3520
    _globals["_EXTRACTIONPOLICY"]._serialized_end = 3735
    _globals["_EXTRACTIONPOLICY_FILTERSENTRY"]._serialized_start = 3689
    _globals["_EXTRACTIONPOLICY_FILTERSENTRY"]._serialized_end = 3735
    _globals["_EXTRACTIONPOLICYREQUEST"]._serialized_start = 3737
    _globals["_EXTRACTIONPOLICYREQUEST"]._serialized_end = 3857
    _globals["_EXTRACTIONPOLICYRESPONSE"]._serialized_start = 3860
    _globals["_EXTRACTIONPOLICYRESPONSE"]._serialized_end = 4303
    _globals[
        "_EXTRACTIONPOLICYRESPONSE_INDEXNAMETABLEMAPPINGENTRY"
    ]._serialized_start = 4180
    _globals["_EXTRACTIONPOLICYRESPONSE_INDEXNAMETABLEMAPPINGENTRY"]._serialized_end = (
        4240
    )
    _globals[
        "_EXTRACTIONPOLICYRESPONSE_OUTPUTINDEXNAMEMAPPINGENTRY"
    ]._serialized_start = 4242
    _globals[
        "_EXTRACTIONPOLICYRESPONSE_OUTPUTINDEXNAMEMAPPINGENTRY"
    ]._serialized_end = 4303
    _globals["_CONTENTMETADATA"]._serialized_start = 4306
    _globals["_CONTENTMETADATA"]._serialized_end = 4597
    _globals["_CONTENTMETADATA_LABELSENTRY"]._serialized_start = 4552
    _globals["_CONTENTMETADATA_LABELSENTRY"]._serialized_end = 4597
    _globals["_CREATECONTENTREQUEST"]._serialized_start = 4599
    _globals["_CREATECONTENTREQUEST"]._serialized_end = 4677
    _globals["_CREATECONTENTRESPONSE"]._serialized_start = 4679
    _globals["_CREATECONTENTRESPONSE"]._serialized_end = 4714
    _globals["_NAMESPACE"]._serialized_start = 4716
    _globals["_NAMESPACE"]._serialized_end = 4799
    _globals["_GETSCHEMAREQUEST"]._serialized_start = 4801
    _globals["_GETSCHEMAREQUEST"]._serialized_end = 4862
    _globals["_GETSCHEMARESPONSE"]._serialized_start = 4864
    _globals["_GETSCHEMARESPONSE"]._serialized_end = 4943
    _globals["_STRUCTUREDDATASCHEMA"]._serialized_start = 4945
    _globals["_STRUCTUREDDATASCHEMA"]._serialized_end = 5008
    _globals["_GETALLSCHEMAREQUEST"]._serialized_start = 5010
    _globals["_GETALLSCHEMAREQUEST"]._serialized_end = 5050
    _globals["_GETALLSCHEMARESPONSE"]._serialized_start = 5052
    _globals["_GETALLSCHEMARESPONSE"]._serialized_end = 5135
    _globals["_COORDINATORSERVICE"]._serialized_start = 5191
    _globals["_COORDINATORSERVICE"]._serialized_end = 7385
# @@protoc_insertion_point(module_scope)



from typing import List, Union, Dict
from .base_extractor import Content, ExtractorWrapper, Feature
from pydantic import Json, BaseModel
import concurrent
import asyncio


class ExtractorModule(BaseModel):
    module_name: str
    class_name: str


def create_extractor_wrapper(extractor_module: ExtractorModule) -> ExtractorWrapper:
    print("creating extractor wrapper")
    global extractor_wrapper
    extractor_wrapper = ExtractorWrapper(
        extractor_module.module_name, extractor_module.class_name
    )


def create_executor(extractor_module: ExtractorModule, workers: int):
    print("creating executor")
    return concurrent.futures.ProcessPoolExecutor(
        initializer=create_extractor_wrapper,
        max_workers=workers,
        initargs=(extractor_module,),
    )


def _extract_content(content_list: Dict[str, Content], params: Json) -> List[Content]:
    return extractor_wrapper.extract_batch(content_list, params)


def _describe():
    return extractor_wrapper.describe()


async def extract_content(
    loop, executor, content_list: Dict[str, Content], params: Json
) -> Dict[str, List[Union[Feature, Content]]]:
    return await loop.run_in_executor(executor, _extract_content, content_list, params)


async def describe(loop, executor):
    return await loop.run_in_executor(executor, _describe)



from abc import abstractmethod
from typing import Any, Callable, List, Literal, Type, Union
from langchain import text_splitter
from pydantic import BaseModel

from indexify_extractor_sdk.base_extractor import (
    Content,
    Extractor,
    Feature,
)


class BaseEmbeddingExtractor(Extractor):
    input_mimes = ["text/plain"]

    def __init__(self, max_context_length: int):
        self._model_context_length: int = max_context_length

    def extract(self, content: Content, params = None) -> List[Union[Feature, Content]]:
        text = content.data.decode("utf-8")
        embedding_list = self.extract_embeddings([text])
        if len(embedding_list) == 0:
            return []
        embedding = embedding_list[0]
        return [Feature.embedding(values=embedding)]

    def extract_batch(self, content_list: List[Content], params  = None) -> List[List[Union[Feature, Content]]]:
        out = []
        texts = [content.data.decode("utf-8") for content in content_list]
        embeddings_list = self.extract_embeddings(texts)
        for embedding in embeddings_list:
            feature = Feature.embedding(values=embedding)
            out.append([feature])
        return out

    @abstractmethod
    def extract_embeddings(self, texts: List[str]) -> List[List[float]]: ...

    def sample_input(self) -> Content:
        return Content.from_text("hello world")



import typer
from . import indexify_extractor, version
from .packager import ExtractorPackager
from typing import Optional
import logging
import os

import sys

import multiprocessing
from typing_extensions import Annotated

cpu_count = multiprocessing.cpu_count()

sys.path.append(".")

extractors_path = os.path.join(os.path.expanduser("~"), ".indexify-extractors")
if os.path.exists(extractors_path):
    sys.path.append(extractors_path)

typer_app = typer.Typer(
    help="indexify-extractor - CLI for running and packaging indexify extractors"
)


# Hack to get around buffered output when not run using interactive terminal in docker
# and to ensure that print statements are flushed immediately
class Unbuffered(object):
    def __init__(self, stream):
        self.stream = stream

    def write(self, data):
        self.stream.write(data)
        self.stream.flush()

    def writelines(self, datas):
        self.stream.writelines(datas)
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)


sys.stdout = Unbuffered(sys.stdout)


def print_version():
    print(f"indexify-extractor-sdk version {version.__version__}")


@typer_app.command()
def describe(extractor: str):
    indexify_extractor.describe_sync(extractor)


@typer_app.command()
def local(
    extractor: str = typer.Argument(
        None,
        help="The extractor name in the format 'module_name:class_name'. For example, 'mock_extractor:MockExtractor'.",
    ),
    text: Optional[str] = None,
    file: Optional[str] = None,
):
    indexify_extractor.local(extractor, text, file)


@typer_app.command()
def join(
    # optional, default to $EXTRACTOR_PATH if not provided. If $EXTRACTOR_PATH is not set, it will raise an error.
    extractor: str = typer.Argument(
        None,
        help="The extractor name in the format 'module_name:class_name'. For example, 'mock_extractor:MockExtractor'.",
    ),
    coordinator_addr: str = "localhost:8950",
    ingestion_addr: str = "localhost:8900",
    workers: Annotated[
        int, typer.Option(help="number of worker processes for extraction")
    ] = cpu_count,
):
    print_version()
    if not extractor:
        extractor = os.environ.get("EXTRACTOR_PATH")
        print(f"Using extractor path from $EXTRACTOR_PATH: {extractor}")
        assert extractor, "Extractor path not provided and $EXTRACTOR_PATH not set."

    print("workers ", workers)

    indexify_extractor.join(extractor, workers, coordinator_addr, ingestion_addr)


@typer_app.command()
def package(
    extractor: str = typer.Argument(
        ...,
        help="The extractor name in the format 'module_name:class_name'. For example, 'mock_extractor:MockExtractor'.",
    ),
    verbose: bool = typer.Option(False, "--verbose", help="Run in verbose mode."),
    dev: bool = typer.Option(False, "--dev", help="Run in development mode."),
    gpu: bool = typer.Option(False, "--gpu", help="Use GPU acceleration."),
):
    """
    Packages an extractor into a Docker image, including Dockerfile generation and tarball creation.
    """
    print_version()
    module_name, class_name = indexify_extractor.split_validate_extractor(extractor)
    logging.basicConfig(level=logging.DEBUG if verbose else logging.INFO)
    packager = ExtractorPackager(
        module_name=module_name,
        class_name=class_name,
        verbose=verbose,
        dev=dev,
        gpu=gpu,
    )
    packager.package()


@typer_app.command()
def download(extractor_path: str = typer.Argument(..., help="Extractor Name")):
    from .downloader import download_extractor

    download_extractor(extractor_path)



import fsspec
import os
import ast
import subprocess
from rich.console import Console
from rich.panel import Panel

console = Console()


class ClassVisitor(ast.NodeVisitor):
    def __init__(self):
        self.classes = []

    def visit_ClassDef(self, node):
        for base in node.bases:
            if isinstance(base, ast.Name) and base.id in ['Extractor', 'BaseEmbeddingExtractor']:
                self.classes.append(node.name)
        self.generic_visit(node)

def find_extractor_subclasses(root_dir):
    for filename in os.listdir(root_dir):
        if filename.endswith('.py'):
            with open(os.path.join(root_dir, filename), 'r') as file:
                content = file.read()
                try:
                    tree = ast.parse(content)
                    visitor = ClassVisitor()
                    visitor.visit(tree)
                    if visitor.classes:
                        base_name = os.path.splitext(filename)[0]
                        return f"{base_name}:{visitor.classes[0]}"
                except SyntaxError as e:
                    print(f"Syntax error in {filename}: {e}")


def print_instructions(directory_path):
    venv_path = os.path.join(directory_path, "ve")

    message = """To run extractor, run the following:\n[bold #4AA4F4]"""

    if not os.environ.get("VIRTUAL_ENV"):
        message += f"source {venv_path}/bin/activate\n"

    message += f"indexify-extractor join {os.path.basename(directory_path)}.{find_extractor_subclasses(directory_path)}[/]"
    console.print(Panel(message, title="[bold magenta]Run the extractor[/]", expand=True))
    
    
def install_dependencies(directory_path):
    console.print("[bold #4AA4F4]Installing dependencies...[/]")
    venv_path = os.path.join(directory_path, "ve")
    requirements_path = os.path.join(directory_path, "requirements.txt")
    
    if not os.path.exists(requirements_path):
            raise ValueError("Unable to find requirements.txt")
        
    if os.environ.get("VIRTUAL_ENV"):
        # install requirements to current env
        subprocess.check_call([os.path.join(os.environ.get("VIRTUAL_ENV"), 'bin', 'pip'), 'install', '--no-deps', '-r', requirements_path])
    else:
        # create env and install requirements
        print("Creating virtual environment...")
        subprocess.check_call(['virtualenv', '-p', "python3.11", venv_path])
        pip_path = os.path.join(venv_path, 'bin', 'pip')

        subprocess.check_call([pip_path, 'install', '--no-deps', '-r', requirements_path])
        subprocess.check_call([pip_path, 'install', 'indexify-extractor-sdk'])

    # print instructions for next steps
    print_instructions(directory_path)
        

def download_extractor(extractor_path):
    console.print("[bold #4AA4F4]Downloading Extractor...[/]")
    extractor_path = extractor_path.removeprefix("hub://")
    fs = fsspec.filesystem("github", org="tensorlakeai", repo="indexify-extractors")
    directory_path = os.path.join(
        os.path.expanduser("~"), ".indexify-extractors", os.path.basename(extractor_path)
    )
    fs.get(extractor_path, directory_path, recursive=True)
    install_dependencies(directory_path)




from indexify_extractor_sdk.base_extractor import Extractor, Content, Feature

from typing import List, Tuple

import json
from pydantic import BaseModel


class InputParams(BaseModel):
    a: int = 0
    b: str = ""


class MockExtractor(Extractor):
    name = "mock_extractor"
    input_mime_types = ["text/plain", "application/pdf", "image/jpeg"]

    system_dependencies = ["sl", "cowsay"]  # some really smalll packages for testing
    python_dependencies = ["tinytext", "pyfiglet"]

    def __init__(self):
        super().__init__()

    def extract(self, content: Content, params: InputParams) -> List[Content]:
        return [
            Content.from_text(
                text="Hello World",
                features=[
                    Feature.embedding(values=[1, 2, 3]),
                    Feature.metadata(json.dumps({"a": 1, "b": "foo"})),
                ],
                labels={"url": "test.com"},
            ),
            Content.from_text(
                text="Pipe Baz",
                features=[Feature.embedding(values=[1, 2, 3])],
                labels={"url": "test.com"},
            ),
        ]

    def sample_input(self) -> Tuple[Content, InputParams]:
        return (
            Content.from_text("hello world"),
            InputParams(a=5, b="h").model_dump_json(),
        )


class MockExtractorsReturnsFeature(Extractor):
    def __init__(self):
        super().__init__()

    def extract(self, content: Content, params: InputParams) -> List[Feature]:
        return [
            Feature.embedding(values=[1, 2, 3]),
            Feature.metadata(json.loads('{"a": 1, "b": "foo"}')),
        ]

    def sample_input(self) -> Tuple[Content, InputParams]:
        return (Content.from_text("hello world"), InputParams(a=5, b="h"))


class MockExtractorNoInputParams(Extractor):
    def __init__(self):
        super().__init__()

    def extract(self, content: Content, params: InputParams) -> List[Content]:
        return [
            Content.from_text(
                text="Hello World", features=[Feature.embedding(values=[1, 2, 3])]
            ),
            Content.from_text(
                text="Pipe Baz", features=[Feature.embedding(values=[1, 2, 3])]
            ),
        ]

    def sample_input(self) -> Content:
        return Content.from_text("hello world")



import asyncio
import gzip
import io
from pathlib import Path
import tarfile
from docker import DockerClient, errors as docker_err
from .packager_utils import (
    DockerfileTemplate,
    ExtractorPathWrapper,
    DynamicModuleLoader,
    async_docker_build,
)
import docker
import logging
import importlib.resources as pkg_resources
import pathlib


class ExtractorPackager:
    """
    Manages the packaging of an extractor into a Docker image, including Dockerfile generation and tarball creation.

    Attributes:
        config: Configuration for the packager.
        docker_client (DockerClient): Client for interacting with Docker.
        logger (Logger): Logger for the packager.

    Methods:
        package: Orchestrates the packaging process, including Dockerfile generation, tarball creation, and Docker image building.
        _generate_dockerfile: Generates the Dockerfile content based on the configuration.
        _generate_compressed_tarball: Creates a compressed tarball containing the Dockerfile and any additional required files.
        _add_dev_dependencies: Adds development dependencies to the tarball if applicable.
    """

    def __init__(
        self,
        module_name: str,
        class_name: str,
        verbose: bool = False,
        dev: bool = False,
        gpu: bool = False,
    ):
        self.logger = logging.getLogger(__name__)
        self.config = {
            "module_name": module_name,
            "class_name": class_name,
            "verbose": verbose,
            "dev": dev,
            "gpu": gpu,
        }
        self.logger.debug(f"Config: {self.config}")

        self.docker_client = DockerClient.from_env()

        self.extractor_path = ExtractorPathWrapper(
            self.config["module_name"], self.config["class_name"]
        ).validate()

        self.extractor_module = self._load_extractor_module()
        self.extractor_description = self._extract_validate_extractor_description()

        if self.config.get("dev", False):
            self.dev_files = self._collect_dev_files()

    def _load_extractor_module(self):
        """
        Dynamically loads the extractor module based on configuration.
        """
        file_name = self.config["module_name"].replace(".", "/") + ".py"
        cwd = pathlib.Path.cwd()
        module_path = cwd.joinpath(file_name)
        self.logger.info(f"Loading module from {module_path}")

        try:
            module = DynamicModuleLoader.load_module_from_path(
                self.config["module_name"], module_path
            )
            return module
        except Exception as e:
            self.logger.error(
                f"Failed to load extractor description for {self.extractor_path.format()} from {module_path}: {e}"
            )
            raise

    def _collect_dev_files(self) -> dict:
        # Initialize the structure for package files
        dev_files = {
            "README.md": self._read_file_text(
                Path(__file__).parent.parent / "README.md"
            ),
            "pyproject.toml": self._read_file_text(
                Path(__file__).parent.parent / "pyproject.toml"
            ),
            "indexify_extractor_sdk": {"__files__": []},
        }
        # Recursively add files from the package directory
        self._add_package_files(
            Path(__file__).parent, "indexify_extractor_sdk", dev_files
        )
        return dev_files

    def _read_file_text(self, path):
        # Utility function to read file text
        try:
            return path.read_text()
        except Exception as e:
            print(f"Error reading file {path}: {e}")
            return None

    def _add_package_files(self, package_path, package_name, acc: dict):
        # Recursively add files from a directory, maintaining structure
        # if the package name doesn't exist, create it
        if package_name not in acc:
            acc[package_name] = {"__files__": []}
        for item in package_path.iterdir():
            if item.is_dir():
                if item.name == "__pycache__":
                    continue  # Skip __pycache__
                # Initialize directory structure
                acc[package_name][item.name] = {"__files__": []}
                # Recurse into directory
                self._add_package_files(item, f"{package_name}.{item.name}", acc)
            elif item.suffix in {".py", ".pyi"}:
                # Add Python files to the list
                acc[package_name]["__files__"].append(str(item.name))

    def package(self):
        try:
            dockerfile_content = self._generate_dockerfile()
            self.logger.debug(dockerfile_content)
        except Exception as e:
            self.logger.error(f"Failed to generate Dockerfile: {e}")
            raise

        try:
            compressed_tar_stream = io.BytesIO(
                self._generate_compressed_tarball(dockerfile_content)
            )
        except Exception as e:
            self.logger.error(f"Failed to generate compressed tarball: {e}")
            raise

        self.logger.info(f"Building image {self.extractor_description['name']}...")
        try:
            self._build_image(self.extractor_description["name"], compressed_tar_stream)
        except Exception as e:
            self.logger.error(f"Failed to build image: {e}")

    def _extract_validate_extractor_description(self) -> dict:
        extractor_cls = getattr(self.extractor_module, self.config["class_name"])

        # we don't need to initialize the class, just get the description
        extractor_description = {
            "name": extractor_cls.name,
            "version": extractor_cls.version,
            "system_dependencies": extractor_cls.system_dependencies,
            "python_dependencies": self._get_python_dependencies(),
        }

        assert (
            extractor_description["name"] is not None
        ), "Extractor.name must be defined"
        assert (
            extractor_description["version"] is not None
        ), "Extractor.version must be defined"

        return extractor_description

    def _get_python_dependencies(self):
        # get module path
        module_path = pathlib.Path.cwd() / (
            self.config["module_name"].replace(".", "/") + ".py"
        )

        # check for requirements.txt
        requirements_path = module_path.joinpath(module_path.parent, "requirements.txt")
        if requirements_path.exists():
            with open(requirements_path, "r") as f:
                requirements = f.read()

            return [entry for entry in requirements.split("\n") if entry]

        # check for class dependencies
        extractor_cls = getattr(self.extractor_module, self.config["class_name"])
        if extractor_cls.python_dependencies:
            return extractor_cls.python_dependencies

        # no dependencies found
        return []

    def _build_image(self, tag: str, fileobj: io.BytesIO):
        try:
            image, _ = asyncio.run(
                async_docker_build(
                    self.docker_client,
                    tag=tag,
                    fileobj=fileobj,
                    custom_context=True,
                    encoding="gzip",
                    rm=True,
                    forcerm=True,
                    pull=True,
                )
            )
            self.logger.info(f"Successfully built image {image.tags[0]}")
        except docker_err.BuildError as e:
            self.logger.error(
                f"Failed to build image {self.extractor_description.get('name')}: {e}"
            )
            raise
        except docker_err.APIError as e:
            self.logger.error(f"Docker API Error: {e}")
            raise

    def _generate_dockerfile(self) -> str:
        return (
            DockerfileTemplate()
            .configure(
                extractor_path=self.extractor_path,
                system_dependencies=" ".join(
                    self.extractor_description["system_dependencies"]
                ),
                python_dependencies=" ".join(
                    self.extractor_description["python_dependencies"]
                ),
                additional_pip_flags=(
                    ""
                    if self.config["gpu"]
                    else "--extra-index-url https://download.pytorch.org/whl/cpu"
                ),
                dev=self.config.get("dev", False),
            )
            .render()
        )

    def _generate_compressed_tarball(self, dockerfile_content: str) -> bytes:
        """
        Generates a tarball containing the Dockerfile and any additional files needed for the extractor.
        """
        dockerfile_tar_buffer = io.BytesIO()
        with tarfile.open(
            fileobj=dockerfile_tar_buffer, mode="w|", format=tarfile.GNU_FORMAT
        ) as tar:
            dockerfile_info = tarfile.TarInfo("Dockerfile")
            dockerfile_info.mode = 0o755
            dockerfile_bytes = dockerfile_content.encode("utf-8")
            dockerfile_info.size = len(dockerfile_bytes)
            tar.addfile(dockerfile_info, fileobj=io.BytesIO(dockerfile_bytes))

            parent_dir = pathlib.Path(self.extractor_path.file_name()).parent

            add_directory_to_tar(tar, parent_dir)

            if self.config.get("dev", False):
                self._add_dev_dependencies(tar)

            # if self.config.verbose, print the directory structure of the tar
            # using tar.list(verbose=True)
            if self.config["verbose"]:
                self.logger.debug("Tarball contents:")
                tar.list(verbose=True)

        dockerfile_tar_buffer.seek(0)
        compressed_data_buffer = io.BytesIO()
        with gzip.GzipFile(fileobj=compressed_data_buffer, mode="wb") as gz:
            gz.write(dockerfile_tar_buffer.getvalue())

        compressed_data_buffer.seek(0)
        return compressed_data_buffer.getvalue()

    # Function to recursively add files from a directory
    def _add_files_from_dir(self, base_path, base_tar_path, tar):
        for file_name, content in self.dev_files[base_path].items():
            if file_name == "__files__":
                for f in content:
                    file_path = Path(base_path) / f
                    file_info = tarfile.TarInfo(name=str(Path(base_tar_path) / f))
                    file_content = pkg_resources.read_text(base_path, str(f)).encode(
                        "utf-8"
                    )
                    file_info.size = len(file_content)
                    file_info.mode = 0o644  # Regular file permissions
                    tar.addfile(file_info, io.BytesIO(file_content))
            else:
                # It's a directory, add it and its contents
                dir_tar_path = str(Path(base_tar_path) / file_name)
                dir_info = tarfile.TarInfo(name=dir_tar_path)
                dir_info.type = tarfile.DIRTYPE
                dir_info.mode = 0o755  # Directory permissions
                tar.addfile(dir_info)
                self._add_files_from_dir(f"{base_path}.{file_name}", dir_tar_path, tar)

    def _add_dev_dependencies(self, tar: tarfile.TarFile):
        # Add the pyproject.toml file
        pyproject_content = self.dev_files["pyproject.toml"].encode("utf-8")
        pyproject_info = tarfile.TarInfo(name="pyproject.toml")
        pyproject_info.size = len(pyproject_content)
        pyproject_info.mode = 0o644  # Regular file permissions
        tar.addfile(pyproject_info, io.BytesIO(pyproject_content))

        # Add the README.md file
        readme_content = self.dev_files["README.md"].encode("utf-8")
        readme_info = tarfile.TarInfo(name="README.md")
        readme_info.size = len(readme_content)
        readme_info.mode = 0o644
        tar.addfile(readme_info, io.BytesIO(readme_content))

        self._add_files_from_dir(
            "indexify_extractor_sdk", "indexify_extractor_sdk", tar
        )


def add_directory_to_tar(tar_file: tarfile.TarFile, directory_path):
    import os

    """
    Add a directory to a tar file with path names relative to the parent directory.

    Parameters:
        directory_path (str): Path to the directory to be added to the tar file.
        tar_file_path (str): Path to the tar file.

    Returns:
        None
    """
    # Iterate over all files and subdirectories in the given directory
    print(f"Adding {directory_path} to tar file")
    for root, dirs, files in os.walk(directory_path):
        for file in files:
            file_path = os.path.join(root, file)
            relative_path = os.path.relpath(file_path, os.path.dirname(directory_path))
            print(f"Adding {file_path} to tar file with relative path {relative_path}")
            tar_file.add(file_path, arcname=relative_path)



import asyncio
from . import coordinator_service_pb2
from .coordinator_service_pb2_grpc import CoordinatorServiceStub
import grpc
from typing import List, Dict, Union
from .base_extractor import Content, Feature
from .content_downloader import download_content, create_content
from .extractor_worker import extract_content
from .ingestion_api_models import (
    ApiContent,
    ApiFeature,
    BeginExtractedContentIngest,
    ExtractedContent,
    ExtractedFeatures,
    FinishExtractedContentIngest,
    ApiBeginExtractedContentIngest,
    ApiExtractedContent,
    ApiExtractedFeatures,
    ApiFinishExtractedContentIngest,
)
from .utils import batched
from .server import http_server, ServerRouter, get_server_advertise_addr
import concurrent
import websockets
from .task_store import TaskStore, CompletedTask

CONTENT_UPLOAD_BATCH_SIZE = 5


class ExtractorAgent:
    def __init__(
        self,
        executor_id: str,
        extractor: coordinator_service_pb2.Extractor,
        coordinator_addr: str,
        executor: concurrent.futures.ProcessPoolExecutor,
        ingestion_addr: str = "localhost:8900",
    ):
        self._task_store: TaskStore = TaskStore()
        self._executor_id = executor_id
        self._extractor = extractor
        self._has_registered = False
        self._coordinator_addr = coordinator_addr
        self._ingestion_addr = ingestion_addr
        self._executor = executor

    async def ticker(self):
        while True:
            await asyncio.sleep(5)
            yield coordinator_service_pb2.HeartbeatRequest(
                executor_id=self._executor_id,
                pending_tasks=self._task_store.num_pending_tasks(),
            )

    async def register(self):
        # This needs to be here because every time we register we need to create a new channel
        # because the old one might have been closed after hb was broken or we could never connect
        self._channel = grpc.aio.insecure_channel(self._coordinator_addr)
        self._stub: CoordinatorServiceStub = CoordinatorServiceStub(self._channel)
        req = coordinator_service_pb2.RegisterExecutorRequest(
            executor_id=self._executor_id,
            addr=self._advertise_addr,
            extractor=self._extractor,
        )
        return await self._stub.RegisterExecutor(req)

    async def task_completion_reporter(self):
        print("starting task completion reporter")
        while True:
            await asyncio.sleep(5)
            await self.launch_tasks()
            # We should copy only the keys and not the values
            url = f"ws://{self._ingestion_addr}/write_content"
            for task_outcome in self._task_store.task_outcomes():
                print(
                    f"reporting outcome of task {task_outcome.task_id}, outcome: {task_outcome.task_outcome}, num_content: {len(task_outcome.new_content)}, num_features: {len(task_outcome.features)}"
                )
                task: coordinator_service_pb2.Task = self._task_store.get_task(
                    task_outcome.task_id
                )
                begin_msg = ApiBeginExtractedContentIngest(
                    BeginExtractedContentIngest=BeginExtractedContentIngest(
                        task_id=task_outcome.task_id,
                        namespace=task.namespace,
                        output_to_index_table_mapping=task.output_index_mapping,
                        parent_content_id=task.content_metadata.id,
                        executor_id=self._executor_id,
                        task_outcome=task_outcome.task_outcome,
                        extraction_policy=task.extraction_policy,
                        extractor=task.extractor,
                    )
                )
                try:
                    async with websockets.connect(
                        url, ping_interval=5, ping_timeout=30
                    ) as ws:
                        await ws.send(begin_msg.model_dump_json())
                        for batch in batched(
                            task_outcome.new_content, CONTENT_UPLOAD_BATCH_SIZE
                        ):
                            extracted_content = ApiExtractedContent(
                                ExtractedContent=ExtractedContent(content_list=batch)
                            )
                            await ws.send(extracted_content.model_dump_json())
                        print(
                            f"finished message {FinishExtractedContentIngest(num_extracted_content=len(task_outcome.new_content)).model_dump_json()}"
                        )
                        for batch in batched(
                            task_outcome.features, CONTENT_UPLOAD_BATCH_SIZE
                        ):
                            extracted_features = ApiExtractedFeatures(
                                ExtractedFeatures=ExtractedFeatures(
                                    content_id=task.content_metadata.id, features=batch
                                )
                            )
                            await ws.send(extracted_features.model_dump_json())
                        finish_msg = ApiFinishExtractedContentIngest(
                            FinishExtractedContentIngest=FinishExtractedContentIngest(
                                num_extracted_content=len(task_outcome.new_content)
                            )
                        )
                        await ws.send(finish_msg.model_dump_json())
                except Exception as e:
                    print(
                        f"failed to report task {task_outcome.task_outcome}, exception: {e}"
                    )
                    continue

                self._task_store.mark_reported(task_id=task_outcome.task_id)

    async def launch_tasks(self):
        tasks_to_launch = self._task_store.get_runnable_tasks()
        if len(tasks_to_launch) == 0:
            return
        print("launching tasks : ", ",".join(tasks_to_launch.keys()))
        content_list = {}
        content_urls = {}
        for _, task in tasks_to_launch.items():
            content_urls[task.id] = task.content_metadata.storage_url
        content_bytes = await download_content(content_urls)
        for task_id, bytes in content_bytes.items():
            if isinstance(bytes, Exception):
                print(f"failed to download content{bytes} for task {task_id}")
                completed_task = CompletedTask(
                    task_id=task_id, task_outcome="Failed", new_content=[], features=[]
                )
                self._task_store.complete(outcome=completed_task)
                continue
            c = create_content(bytes, tasks_to_launch[task_id])
            content_list[task_id] = c
        if len(content_list) == 0:
            return
        try:
            print(f"launching tasks {len(content_list)}")
            outputs: Dict[str, Union[List[Feature], List[Content]]] = (
                await extract_content(
                    loop=asyncio.get_running_loop(),
                    executor=self._executor,
                    content_list=content_list,
                    params=task.input_params,
                )
            )
        except Exception as e:
            task_ids = ",".join(content_list.keys())
            print(f"failed to execute tasks {task_ids} {e}")
            for task_id in content_list.keys():
                completed_task = CompletedTask(
                    task_id=task_id, task_outcome="Failed", new_content=[], features=[]
                )
                self._task_store.complete(outcome=completed_task)
            return
        print(f"completed task len {len(outputs)}")
        for task_id, e_output in outputs.items():
            print(f"completed task {task_id}")
            new_content: List[ApiContent] = []
            new_features: List[ApiFeature] = []
            out: Union[Feature, Content]
            for out in e_output:
                if type(out) == Feature:
                    new_features.append(ApiFeature.from_feature(feature=out))
                    continue
                new_content.append(ApiContent.from_content(content=out))
            completed_task = CompletedTask(
                task_id=task_id,
                task_outcome="Success",
                new_content=new_content,
                features=new_features,
            )
            self._task_store.complete(outcome=completed_task)

    async def run(self):
        import signal

        asyncio.get_event_loop().add_signal_handler(
            signal.SIGINT, self.shutdown, asyncio.get_event_loop()
        )
        server_router = ServerRouter(self._executor)
        self._http_server = http_server(server_router)
        asyncio.create_task(self._http_server.serve())
        self._advertise_addr = await get_server_advertise_addr(self._http_server)
        print(f"advertise addr is {self._advertise_addr}")
        asyncio.create_task(self.task_completion_reporter())
        self._should_run = True
        while self._should_run:
            print("attempting to register")
            try:
                await self.register()
                self._has_registered = True
            except Exception as e:
                print(f"failed to register{e}")
                await asyncio.sleep(5)
                continue
            hb_ticker = self.ticker()
            print("starting heartbeat")
            try:
                hb_response_it = self._stub.Heartbeat(hb_ticker)
                resp: coordinator_service_pb2.HeartbeatResponse
                async for resp in hb_response_it:
                    self._task_store.add_tasks(resp.tasks)
            except Exception as e:
                print(f"failed to heartbeat{e}")
                continue

    async def _shutdown(self, loop):
        print("shutting down agent ...")
        self._should_run = False
        self._http_server.should_exit = True
        await self._channel.close()
        for task in asyncio.all_tasks(loop):
            task.cancel()

    def shutdown(self, loop):
        self._executor.shutdown(wait=True, cancel_futures=True)
        loop.create_task(self._shutdown(loop))



from pydantic import BaseModel, Json
from typing import Dict, Optional, List
from .ingestion_api_models import ApiContent, ApiFeature
from .base_extractor import Content, Feature
from .extractor_worker import extract_content, ExtractorModule
import uvicorn
import asyncio
import json
import netifaces
import concurrent

from fastapi import FastAPI, APIRouter


class ExtractionRequest(BaseModel):
    content: ApiContent
    input_params: Optional[Json]


class ExtractionResponse(BaseModel):
    content: List[ApiContent]
    features: List[ApiFeature]


class ServerRouter:
    def __init__(self, executor: concurrent.futures.ProcessPoolExecutor):
        self._executor = executor
        self.router = APIRouter()
        self.router.add_api_route("/", self.root, methods=["GET"])
        self.router.add_api_route("/extract", self.extract, methods=["POST"])

    async def root(self):
        return {"Indexify Extractor"}

    async def extract(self, request: ExtractionRequest):
        loop = asyncio.get_event_loop()
        content = Content(
            content_type=request.content.content_type,
            data=bytes(request.content.bytes),
            features=[],
            labels=request.content.labels,
        )
        task_id = "dummy_task_id"
        content_dict: Dict[str, Content] = { task_id: content }
        input_params = (
            json.dumps(request.input_params) if request.input_params else None
        )
        extractor_out: Dict[str, List[Content]] = await extract_content(
            loop, self._executor, content_dict, params=input_params
        )
        api_content: List[ApiContent] = []
        api_features: List[ApiFeature] = []
        
        for out_list in extractor_out.values():
            for out in out_list:
                if type(out) == Feature:
                    api_features.append(ApiFeature.from_feature(out))
                    continue
                api_content.append(ApiContent.from_content(out))
        return ExtractionResponse(content=api_content,features=api_features)


class ServerWithNoSigHandler(uvicorn.Server):
    def install_signal_handlers(self) -> None:
        pass


def http_server(server_router: ServerRouter) -> uvicorn.Server:
    print("starting extraction server endpoint")
    app = FastAPI()
    app.include_router(server_router.router)
    config = uvicorn.Config(
        app, loop="asyncio", host="0.0.0.0", port=0, log_level="info", lifespan="off"
    )

    return ServerWithNoSigHandler(config)


async def get_server_advertise_addr(
    server: uvicorn.Server, advertise_addr: str = ""
) -> str:
    while not server.started:
        await asyncio.sleep(0.1)
    port: int
    for server in server.servers:
        for sock in server.sockets:
            port = sock.getsockname()[1]
            break
    addr = (
        get_most_publicly_addressable_ip() if advertise_addr == "" else advertise_addr
    )
    return f"{addr}:{port}"


def get_most_publicly_addressable_ip():
    # IP address priorities
    ip_priorities = {"public": 1, "private": 2, "loopback": 3}

    def ip_type(ip_address):
        """Determine IP address type based on known ranges."""
        if ip_address.startswith("127."):
            return "loopback"
        elif (
            ip_address.startswith("10.")
            or ip_address.startswith("172.16.")
            or ip_address.startswith("172.31.")
            or ip_address.startswith("192.168.")
        ):
            return "private"
        else:
            return "public"

    def sort_key(ip):
        """Sort key for IP addresses based on their type."""
        return ip_priorities[ip_type(ip)]

    ip_addresses = []
    # List all network interfaces
    for interface in netifaces.interfaces():
        addrs = netifaces.ifaddresses(interface)
        # Consider IPv4 addresses only
        if netifaces.AF_INET in addrs:
            for addr_info in addrs[netifaces.AF_INET]:
                ip_addresses.append(addr_info["addr"])

    # Sort IP addresses by their type
    ip_addresses.sort(key=sort_key)

    # Return the most publicly addressable IP, if available
    if ip_addresses:
        return ip_addresses[0]
    else:
        raise Exception("No IP address available")



from pydantic import BaseModel

from typing import List

from indexify_extractor_sdk import Extractor, Content, Feature
from indexify_extractor_sdk.base_extractor import Content
import json


# Extractors can be parameterized by providing a pydantic model
# as the second argument to the Extractor class. The model is exposed
# in the Indexify API so that users can dynamically change the behavior
# of the extractor for various use cases. Some examples here can be
# chunk size of audio clips during segmentation, or the text splitter algorithm
# for embedding long documents
class InputParams(BaseModel):
    a: int = 0
    b: str = ""


class MyExtractor(Extractor):
    name = "your-docker-hub-username/MyExtractor"
    description = "Description of the extractor goes here."

    # Any python dependencies included in the extractor must be listed here.
    python_dependencies = ["torch", "transformers"]

    # Any system dependencies that the python code here depends on needs to be listed here
    # We use Ubuntu base images, so any ubuntu package can be installed here.
    system_dependencies = []

    # The mime types of content that this extractor can process. Any content ingested into
    # Indexify which does not match one of these mime types will not be sent to this extractor.
    input_mime_types = ["text/plain"]

    def __init__(self):
        super().__init__()

    # The extract method is the main entrypoint for the extractor. It takes in a Content object and returns
    # a list of transformed content objects. Each transformed content can have different features that will be
    # indexed. For example a long document can be split into multiple content chunks, each containing features
    # like embedding, JSON metdata, etc. If you are doing on ETL(transformation), you can return content with
    # no features.
    def extract(self, content: Content, params: InputParams) -> List[Content]:
        return [
            ## If the name of the embedding field in the schema is anything besides "embedding",
            # you must specify the name of the field in the Feature.embedding call.
            # Feature.embedding(value=[1, 2, 3], name="my_embedding")
            Content.from_text(
                text="Hello World", feature=Feature.embedding(values=[1, 2, 3])
            ),
            Content.from_text(
                text="Pipe Baz", feature=Feature.embedding(values=[1, 2, 3])
            ),
            Content.from_text(
                text="Hello World",
                feature=Feature.metadata(value=json.dumps({"key": "value"})),
            ),
        ]

    # Provide some representative sample input that the extractor can process.
    def sample_input(self) -> Content:
        Content.from_text(text="Hello World")


if __name__ == "__main__":
    # You can run the extractor by simply invoking this python file
    # python custom_extractor.py and that would run the extractor
    # with the sample input provided.
    MyExtractor().extract_sample_input()



from pydantic import BaseModel, Json
from typing import List, Dict
import json
from .base_extractor import Feature, Content


class ApiFeature(BaseModel):
    feature_type: str
    name: str
    data: Json

    @classmethod
    def from_feature(cls, feature: Feature):
        return cls(
            feature_type=feature.feature_type,
            name=feature.name,
            data=json.dumps(feature.value),
        )


class ApiContent(BaseModel):
    content_type: str
    bytes: List[int]
    features: List[ApiFeature] = []
    labels: Dict[str, str] = {}

    @classmethod
    def from_content(cls, content: Content):
        content_features = []
        for feature in content.features:
            content_features.append(ApiFeature.from_feature(feature=feature))
        return cls(
            content_type=content.content_type,
            bytes=list(content.data),
            features=content_features,
            labels=content.labels,
        )


class BeginExtractedContentIngest(BaseModel):
    task_id: str
    namespace: str
    output_to_index_table_mapping: Dict[str, str]
    parent_content_id: str
    executor_id: str
    task_outcome: str
    extraction_policy: str
    extractor: str


class ExtractedFeatures(BaseModel):
    content_id: str
    features: List[ApiFeature]


class ExtractedContent(BaseModel):
    content_list: List[ApiContent]


class FinishExtractedContentIngest(BaseModel):
    num_extracted_content: int


class ApiBeginExtractedContentIngest(BaseModel):
    BeginExtractedContentIngest: BeginExtractedContentIngest


class ApiExtractedContent(BaseModel):
    ExtractedContent: ExtractedContent


class ApiFinishExtractedContentIngest(BaseModel):
    FinishExtractedContentIngest: FinishExtractedContentIngest


class ApiExtractedFeatures(BaseModel):
    ExtractedFeatures: ExtractedFeatures



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class TaskOutcome(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    UNKNOWN: _ClassVar[TaskOutcome]
    FAILED: _ClassVar[TaskOutcome]
    SUCCESS: _ClassVar[TaskOutcome]

UNKNOWN: TaskOutcome
FAILED: TaskOutcome
SUCCESS: TaskOutcome

class GetContentMetadataRequest(_message.Message):
    __slots__ = ("content_list",)
    CONTENT_LIST_FIELD_NUMBER: _ClassVar[int]
    content_list: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, content_list: _Optional[_Iterable[str]] = ...) -> None: ...

class GetContentMetadataResponse(_message.Message):
    __slots__ = ("content_list",)
    CONTENT_LIST_FIELD_NUMBER: _ClassVar[int]
    content_list: _containers.RepeatedCompositeFieldContainer[ContentMetadata]
    def __init__(
        self,
        content_list: _Optional[_Iterable[_Union[ContentMetadata, _Mapping]]] = ...,
    ) -> None: ...

class UpdateTaskRequest(_message.Message):
    __slots__ = ("executor_id", "task_id", "outcome", "content_list")
    EXECUTOR_ID_FIELD_NUMBER: _ClassVar[int]
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    OUTCOME_FIELD_NUMBER: _ClassVar[int]
    CONTENT_LIST_FIELD_NUMBER: _ClassVar[int]
    executor_id: str
    task_id: str
    outcome: TaskOutcome
    content_list: _containers.RepeatedCompositeFieldContainer[ContentMetadata]
    def __init__(
        self,
        executor_id: _Optional[str] = ...,
        task_id: _Optional[str] = ...,
        outcome: _Optional[_Union[TaskOutcome, str]] = ...,
        content_list: _Optional[_Iterable[_Union[ContentMetadata, _Mapping]]] = ...,
    ) -> None: ...

class ListStateChangesRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class StateChange(_message.Message):
    __slots__ = ("id", "object_id", "change_type", "created_at", "processed_at")
    ID_FIELD_NUMBER: _ClassVar[int]
    OBJECT_ID_FIELD_NUMBER: _ClassVar[int]
    CHANGE_TYPE_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    PROCESSED_AT_FIELD_NUMBER: _ClassVar[int]
    id: str
    object_id: str
    change_type: str
    created_at: int
    processed_at: int
    def __init__(
        self,
        id: _Optional[str] = ...,
        object_id: _Optional[str] = ...,
        change_type: _Optional[str] = ...,
        created_at: _Optional[int] = ...,
        processed_at: _Optional[int] = ...,
    ) -> None: ...

class ListStateChangesResponse(_message.Message):
    __slots__ = ("changes",)
    CHANGES_FIELD_NUMBER: _ClassVar[int]
    changes: _containers.RepeatedCompositeFieldContainer[StateChange]
    def __init__(
        self, changes: _Optional[_Iterable[_Union[StateChange, _Mapping]]] = ...
    ) -> None: ...

class ListTasksRequest(_message.Message):
    __slots__ = ("namespace", "extraction_policy")
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    EXTRACTION_POLICY_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    extraction_policy: str
    def __init__(
        self, namespace: _Optional[str] = ..., extraction_policy: _Optional[str] = ...
    ) -> None: ...

class ListTasksResponse(_message.Message):
    __slots__ = ("tasks",)
    TASKS_FIELD_NUMBER: _ClassVar[int]
    tasks: _containers.RepeatedCompositeFieldContainer[Task]
    def __init__(
        self, tasks: _Optional[_Iterable[_Union[Task, _Mapping]]] = ...
    ) -> None: ...

class UpdateTaskResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetExtractorCoordinatesRequest(_message.Message):
    __slots__ = ("extractor",)
    EXTRACTOR_FIELD_NUMBER: _ClassVar[int]
    extractor: str
    def __init__(self, extractor: _Optional[str] = ...) -> None: ...

class GetExtractorCoordinatesResponse(_message.Message):
    __slots__ = ("addrs",)
    ADDRS_FIELD_NUMBER: _ClassVar[int]
    addrs: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, addrs: _Optional[_Iterable[str]] = ...) -> None: ...

class ListIndexesRequest(_message.Message):
    __slots__ = ("namespace",)
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    def __init__(self, namespace: _Optional[str] = ...) -> None: ...

class ListIndexesResponse(_message.Message):
    __slots__ = ("indexes",)
    INDEXES_FIELD_NUMBER: _ClassVar[int]
    indexes: _containers.RepeatedCompositeFieldContainer[Index]
    def __init__(
        self, indexes: _Optional[_Iterable[_Union[Index, _Mapping]]] = ...
    ) -> None: ...

class GetIndexRequest(_message.Message):
    __slots__ = ("namespace", "name")
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    name: str
    def __init__(
        self, namespace: _Optional[str] = ..., name: _Optional[str] = ...
    ) -> None: ...

class GetIndexResponse(_message.Message):
    __slots__ = ("index",)
    INDEX_FIELD_NUMBER: _ClassVar[int]
    index: Index
    def __init__(self, index: _Optional[_Union[Index, _Mapping]] = ...) -> None: ...

class CreateIndexRequest(_message.Message):
    __slots__ = ("index",)
    INDEX_FIELD_NUMBER: _ClassVar[int]
    index: Index
    def __init__(self, index: _Optional[_Union[Index, _Mapping]] = ...) -> None: ...

class CreateIndexResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class Index(_message.Message):
    __slots__ = (
        "name",
        "namespace",
        "table_name",
        "schema",
        "extraction_policy",
        "extractor",
    )
    NAME_FIELD_NUMBER: _ClassVar[int]
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    TABLE_NAME_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    EXTRACTION_POLICY_FIELD_NUMBER: _ClassVar[int]
    EXTRACTOR_FIELD_NUMBER: _ClassVar[int]
    name: str
    namespace: str
    table_name: str
    schema: str
    extraction_policy: str
    extractor: str
    def __init__(
        self,
        name: _Optional[str] = ...,
        namespace: _Optional[str] = ...,
        table_name: _Optional[str] = ...,
        schema: _Optional[str] = ...,
        extraction_policy: _Optional[str] = ...,
        extractor: _Optional[str] = ...,
    ) -> None: ...

class Embedding(_message.Message):
    __slots__ = ("embedding",)
    EMBEDDING_FIELD_NUMBER: _ClassVar[int]
    embedding: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, embedding: _Optional[_Iterable[float]] = ...) -> None: ...

class Attributes(_message.Message):
    __slots__ = ("attributes",)
    ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    attributes: str
    def __init__(self, attributes: _Optional[str] = ...) -> None: ...

class Feature(_message.Message):
    __slots__ = ("name", "embedding", "attributes")
    NAME_FIELD_NUMBER: _ClassVar[int]
    EMBEDDING_FIELD_NUMBER: _ClassVar[int]
    ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    name: str
    embedding: Embedding
    attributes: Attributes
    def __init__(
        self,
        name: _Optional[str] = ...,
        embedding: _Optional[_Union[Embedding, _Mapping]] = ...,
        attributes: _Optional[_Union[Attributes, _Mapping]] = ...,
    ) -> None: ...

class Content(_message.Message):
    __slots__ = ("mime", "data", "features")
    MIME_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    FEATURES_FIELD_NUMBER: _ClassVar[int]
    mime: str
    data: bytes
    features: _containers.RepeatedCompositeFieldContainer[Feature]
    def __init__(
        self,
        mime: _Optional[str] = ...,
        data: _Optional[bytes] = ...,
        features: _Optional[_Iterable[_Union[Feature, _Mapping]]] = ...,
    ) -> None: ...

class RegisterExecutorRequest(_message.Message):
    __slots__ = ("executor_id", "addr", "extractor")
    EXECUTOR_ID_FIELD_NUMBER: _ClassVar[int]
    ADDR_FIELD_NUMBER: _ClassVar[int]
    EXTRACTOR_FIELD_NUMBER: _ClassVar[int]
    executor_id: str
    addr: str
    extractor: Extractor
    def __init__(
        self,
        executor_id: _Optional[str] = ...,
        addr: _Optional[str] = ...,
        extractor: _Optional[_Union[Extractor, _Mapping]] = ...,
    ) -> None: ...

class RegisterExecutorResponse(_message.Message):
    __slots__ = ("executor_id",)
    EXECUTOR_ID_FIELD_NUMBER: _ClassVar[int]
    executor_id: str
    def __init__(self, executor_id: _Optional[str] = ...) -> None: ...

class HeartbeatRequest(_message.Message):
    __slots__ = ("executor_id", "pending_tasks")
    EXECUTOR_ID_FIELD_NUMBER: _ClassVar[int]
    PENDING_TASKS_FIELD_NUMBER: _ClassVar[int]
    executor_id: str
    pending_tasks: int
    def __init__(
        self, executor_id: _Optional[str] = ..., pending_tasks: _Optional[int] = ...
    ) -> None: ...

class HeartbeatResponse(_message.Message):
    __slots__ = ("executor_id", "tasks")
    EXECUTOR_ID_FIELD_NUMBER: _ClassVar[int]
    TASKS_FIELD_NUMBER: _ClassVar[int]
    executor_id: str
    tasks: _containers.RepeatedCompositeFieldContainer[Task]
    def __init__(
        self,
        executor_id: _Optional[str] = ...,
        tasks: _Optional[_Iterable[_Union[Task, _Mapping]]] = ...,
    ) -> None: ...

class Task(_message.Message):
    __slots__ = (
        "id",
        "extractor",
        "namespace",
        "content_metadata",
        "input_params",
        "extraction_policy",
        "output_index_mapping",
        "outcome",
    )

    class OutputIndexMappingEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    ID_FIELD_NUMBER: _ClassVar[int]
    EXTRACTOR_FIELD_NUMBER: _ClassVar[int]
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    CONTENT_METADATA_FIELD_NUMBER: _ClassVar[int]
    INPUT_PARAMS_FIELD_NUMBER: _ClassVar[int]
    EXTRACTION_POLICY_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_INDEX_MAPPING_FIELD_NUMBER: _ClassVar[int]
    OUTCOME_FIELD_NUMBER: _ClassVar[int]
    id: str
    extractor: str
    namespace: str
    content_metadata: ContentMetadata
    input_params: str
    extraction_policy: str
    output_index_mapping: _containers.ScalarMap[str, str]
    outcome: TaskOutcome
    def __init__(
        self,
        id: _Optional[str] = ...,
        extractor: _Optional[str] = ...,
        namespace: _Optional[str] = ...,
        content_metadata: _Optional[_Union[ContentMetadata, _Mapping]] = ...,
        input_params: _Optional[str] = ...,
        extraction_policy: _Optional[str] = ...,
        output_index_mapping: _Optional[_Mapping[str, str]] = ...,
        outcome: _Optional[_Union[TaskOutcome, str]] = ...,
    ) -> None: ...

class ListExtractorsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ListExtractorsResponse(_message.Message):
    __slots__ = ("extractors",)
    EXTRACTORS_FIELD_NUMBER: _ClassVar[int]
    extractors: _containers.RepeatedCompositeFieldContainer[Extractor]
    def __init__(
        self, extractors: _Optional[_Iterable[_Union[Extractor, _Mapping]]] = ...
    ) -> None: ...

class Extractor(_message.Message):
    __slots__ = (
        "name",
        "description",
        "input_params",
        "embedding_schemas",
        "metadata_schemas",
        "input_mime_types",
    )

    class EmbeddingSchemasEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    class MetadataSchemasEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    INPUT_PARAMS_FIELD_NUMBER: _ClassVar[int]
    EMBEDDING_SCHEMAS_FIELD_NUMBER: _ClassVar[int]
    METADATA_SCHEMAS_FIELD_NUMBER: _ClassVar[int]
    INPUT_MIME_TYPES_FIELD_NUMBER: _ClassVar[int]
    name: str
    description: str
    input_params: str
    embedding_schemas: _containers.ScalarMap[str, str]
    metadata_schemas: _containers.ScalarMap[str, str]
    input_mime_types: _containers.RepeatedScalarFieldContainer[str]
    def __init__(
        self,
        name: _Optional[str] = ...,
        description: _Optional[str] = ...,
        input_params: _Optional[str] = ...,
        embedding_schemas: _Optional[_Mapping[str, str]] = ...,
        metadata_schemas: _Optional[_Mapping[str, str]] = ...,
        input_mime_types: _Optional[_Iterable[str]] = ...,
    ) -> None: ...

class GetNamespaceRequest(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class GetNamespaceResponse(_message.Message):
    __slots__ = ("namespace",)
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    namespace: Namespace
    def __init__(
        self, namespace: _Optional[_Union[Namespace, _Mapping]] = ...
    ) -> None: ...

class ListContentRequest(_message.Message):
    __slots__ = ("namespace", "source", "parent_id", "labels_eq")

    class LabelsEqEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FIELD_NUMBER: _ClassVar[int]
    PARENT_ID_FIELD_NUMBER: _ClassVar[int]
    LABELS_EQ_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    source: str
    parent_id: str
    labels_eq: _containers.ScalarMap[str, str]
    def __init__(
        self,
        namespace: _Optional[str] = ...,
        source: _Optional[str] = ...,
        parent_id: _Optional[str] = ...,
        labels_eq: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class ListContentResponse(_message.Message):
    __slots__ = ("content_list",)
    CONTENT_LIST_FIELD_NUMBER: _ClassVar[int]
    content_list: _containers.RepeatedCompositeFieldContainer[ContentMetadata]
    def __init__(
        self,
        content_list: _Optional[_Iterable[_Union[ContentMetadata, _Mapping]]] = ...,
    ) -> None: ...

class ListExtractionPoliciesRequest(_message.Message):
    __slots__ = ("namespace",)
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    def __init__(self, namespace: _Optional[str] = ...) -> None: ...

class ListExtractionPoliciesResponse(_message.Message):
    __slots__ = ("policies",)
    POLICIES_FIELD_NUMBER: _ClassVar[int]
    policies: _containers.RepeatedCompositeFieldContainer[ExtractionPolicy]
    def __init__(
        self, policies: _Optional[_Iterable[_Union[ExtractionPolicy, _Mapping]]] = ...
    ) -> None: ...

class CreateNamespaceRequest(_message.Message):
    __slots__ = ("name", "policies")
    NAME_FIELD_NUMBER: _ClassVar[int]
    POLICIES_FIELD_NUMBER: _ClassVar[int]
    name: str
    policies: _containers.RepeatedCompositeFieldContainer[ExtractionPolicy]
    def __init__(
        self,
        name: _Optional[str] = ...,
        policies: _Optional[_Iterable[_Union[ExtractionPolicy, _Mapping]]] = ...,
    ) -> None: ...

class CreateNamespaceResponse(_message.Message):
    __slots__ = ("name", "created_at")
    NAME_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    name: str
    created_at: int
    def __init__(
        self, name: _Optional[str] = ..., created_at: _Optional[int] = ...
    ) -> None: ...

class ListNamespaceRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ListNamespaceResponse(_message.Message):
    __slots__ = ("namespaces",)
    NAMESPACES_FIELD_NUMBER: _ClassVar[int]
    namespaces: _containers.RepeatedCompositeFieldContainer[Namespace]
    def __init__(
        self, namespaces: _Optional[_Iterable[_Union[Namespace, _Mapping]]] = ...
    ) -> None: ...

class ExtractionPolicy(_message.Message):
    __slots__ = ("extractor", "name", "input_params", "filters", "content_source")

    class FiltersEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    EXTRACTOR_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    INPUT_PARAMS_FIELD_NUMBER: _ClassVar[int]
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    CONTENT_SOURCE_FIELD_NUMBER: _ClassVar[int]
    extractor: str
    name: str
    input_params: str
    filters: _containers.ScalarMap[str, str]
    content_source: str
    def __init__(
        self,
        extractor: _Optional[str] = ...,
        name: _Optional[str] = ...,
        input_params: _Optional[str] = ...,
        filters: _Optional[_Mapping[str, str]] = ...,
        content_source: _Optional[str] = ...,
    ) -> None: ...

class ExtractionPolicyRequest(_message.Message):
    __slots__ = ("namespace", "policy", "created_at")
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    POLICY_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    policy: ExtractionPolicy
    created_at: int
    def __init__(
        self,
        namespace: _Optional[str] = ...,
        policy: _Optional[_Union[ExtractionPolicy, _Mapping]] = ...,
        created_at: _Optional[int] = ...,
    ) -> None: ...

class ExtractionPolicyResponse(_message.Message):
    __slots__ = (
        "created_at",
        "extractor",
        "index_name_table_mapping",
        "output_index_name_mapping",
    )

    class IndexNameTableMappingEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    class OutputIndexNameMappingEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    EXTRACTOR_FIELD_NUMBER: _ClassVar[int]
    INDEX_NAME_TABLE_MAPPING_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_INDEX_NAME_MAPPING_FIELD_NUMBER: _ClassVar[int]
    created_at: int
    extractor: Extractor
    index_name_table_mapping: _containers.ScalarMap[str, str]
    output_index_name_mapping: _containers.ScalarMap[str, str]
    def __init__(
        self,
        created_at: _Optional[int] = ...,
        extractor: _Optional[_Union[Extractor, _Mapping]] = ...,
        index_name_table_mapping: _Optional[_Mapping[str, str]] = ...,
        output_index_name_mapping: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class ContentMetadata(_message.Message):
    __slots__ = (
        "id",
        "file_name",
        "parent_id",
        "mime",
        "labels",
        "storage_url",
        "created_at",
        "namespace",
        "source",
        "size_bytes",
    )

    class LabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    ID_FIELD_NUMBER: _ClassVar[int]
    FILE_NAME_FIELD_NUMBER: _ClassVar[int]
    PARENT_ID_FIELD_NUMBER: _ClassVar[int]
    MIME_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    STORAGE_URL_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FIELD_NUMBER: _ClassVar[int]
    SIZE_BYTES_FIELD_NUMBER: _ClassVar[int]
    id: str
    file_name: str
    parent_id: str
    mime: str
    labels: _containers.ScalarMap[str, str]
    storage_url: str
    created_at: int
    namespace: str
    source: str
    size_bytes: int
    def __init__(
        self,
        id: _Optional[str] = ...,
        file_name: _Optional[str] = ...,
        parent_id: _Optional[str] = ...,
        mime: _Optional[str] = ...,
        labels: _Optional[_Mapping[str, str]] = ...,
        storage_url: _Optional[str] = ...,
        created_at: _Optional[int] = ...,
        namespace: _Optional[str] = ...,
        source: _Optional[str] = ...,
        size_bytes: _Optional[int] = ...,
    ) -> None: ...

class CreateContentRequest(_message.Message):
    __slots__ = ("content",)
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    content: ContentMetadata
    def __init__(
        self, content: _Optional[_Union[ContentMetadata, _Mapping]] = ...
    ) -> None: ...

class CreateContentResponse(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class Namespace(_message.Message):
    __slots__ = ("name", "policies")
    NAME_FIELD_NUMBER: _ClassVar[int]
    POLICIES_FIELD_NUMBER: _ClassVar[int]
    name: str
    policies: _containers.RepeatedCompositeFieldContainer[ExtractionPolicy]
    def __init__(
        self,
        name: _Optional[str] = ...,
        policies: _Optional[_Iterable[_Union[ExtractionPolicy, _Mapping]]] = ...,
    ) -> None: ...

class GetSchemaRequest(_message.Message):
    __slots__ = ("namespace", "content_source")
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    CONTENT_SOURCE_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    content_source: str
    def __init__(
        self, namespace: _Optional[str] = ..., content_source: _Optional[str] = ...
    ) -> None: ...

class GetSchemaResponse(_message.Message):
    __slots__ = ("schema",)
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    schema: StructuredDataSchema
    def __init__(
        self, schema: _Optional[_Union[StructuredDataSchema, _Mapping]] = ...
    ) -> None: ...

class StructuredDataSchema(_message.Message):
    __slots__ = ("columns", "content_source")
    COLUMNS_FIELD_NUMBER: _ClassVar[int]
    CONTENT_SOURCE_FIELD_NUMBER: _ClassVar[int]
    columns: str
    content_source: str
    def __init__(
        self, columns: _Optional[str] = ..., content_source: _Optional[str] = ...
    ) -> None: ...

class GetAllSchemaRequest(_message.Message):
    __slots__ = ("namespace",)
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    def __init__(self, namespace: _Optional[str] = ...) -> None: ...

class GetAllSchemaResponse(_message.Message):
    __slots__ = ("schemas",)
    SCHEMAS_FIELD_NUMBER: _ClassVar[int]
    schemas: _containers.RepeatedCompositeFieldContainer[StructuredDataSchema]
    def __init__(
        self,
        schemas: _Optional[_Iterable[_Union[StructuredDataSchema, _Mapping]]] = ...,
    ) -> None: ...



from . import coordinator_service_pb2

from typing import Dict, List

from pydantic import BaseModel
from .ingestion_api_models import ApiContent, ApiFeature


class CompletedTask(BaseModel):
    task_id: str
    task_outcome: str
    new_content: List[ApiContent]
    features: List[ApiFeature]


class TaskStore:
    def __init__(self) -> None:
        self._tasks: Dict[str, coordinator_service_pb2.Task] = {}
        self._running_tasks: Dict[str, coordinator_service_pb2.Task] = {}
        self._finished: Dict[str, CompletedTask] = {}

    def get_task(self, id) -> coordinator_service_pb2.Task:
        return self._tasks[id]

    def add_tasks(self, tasks: List[coordinator_service_pb2.Task]):
        for task in tasks:
            if (
                (task.id in self._tasks)
                or (task.id in self._running_tasks)
                or (task.id in self._finished)
            ):
                continue
            print(f"added task {task.id} to queue")
            self._tasks[task.id] = task

    def get_runnable_tasks(self) -> Dict[str, coordinator_service_pb2.Task]:
        runnable_tasks = set(self._tasks) - set(self._running_tasks)
        runnable_tasks = set(runnable_tasks) - set(self._finished)
        out = {}
        for task_id in runnable_tasks:
            out[task_id] = self._tasks[task_id]
            self._running_tasks[task_id] = self._tasks[task_id]
        return out

    def complete(self, outcome: CompletedTask):
        self._finished[outcome.task_id] = outcome
        if outcome.task_id in self._running_tasks:
            self._running_tasks.pop(outcome.task_id)

    def mark_reported(self, task_id: str):
        self._tasks.pop(task_id)
        self._finished.pop(task_id)

    def num_pending_tasks(self) -> int:
        return len(self._tasks) + len(self._running_tasks)

    def task_outcomes(self) -> List[CompletedTask]:
        return self._finished.copy().values()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import coordinator_service_pb2 as coordinator__service__pb2


class CoordinatorServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreateContent = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/CreateContent",
            request_serializer=coordinator__service__pb2.CreateContentRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.CreateContentResponse.FromString,
        )
        self.GetContentMetadata = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/GetContentMetadata",
            request_serializer=coordinator__service__pb2.GetContentMetadataRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.GetContentMetadataResponse.FromString,
        )
        self.ListContent = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/ListContent",
            request_serializer=coordinator__service__pb2.ListContentRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.ListContentResponse.FromString,
        )
        self.CreateExtractionPolicy = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/CreateExtractionPolicy",
            request_serializer=coordinator__service__pb2.ExtractionPolicyRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.ExtractionPolicyResponse.FromString,
        )
        self.ListExtractionPolicies = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/ListExtractionPolicies",
            request_serializer=coordinator__service__pb2.ListExtractionPoliciesRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.ListExtractionPoliciesResponse.FromString,
        )
        self.CreateNS = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/CreateNS",
            request_serializer=coordinator__service__pb2.CreateNamespaceRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.CreateNamespaceResponse.FromString,
        )
        self.ListNS = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/ListNS",
            request_serializer=coordinator__service__pb2.ListNamespaceRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.ListNamespaceResponse.FromString,
        )
        self.GetNS = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/GetNS",
            request_serializer=coordinator__service__pb2.GetNamespaceRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.GetNamespaceResponse.FromString,
        )
        self.ListExtractors = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/ListExtractors",
            request_serializer=coordinator__service__pb2.ListExtractorsRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.ListExtractorsResponse.FromString,
        )
        self.RegisterExecutor = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/RegisterExecutor",
            request_serializer=coordinator__service__pb2.RegisterExecutorRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.RegisterExecutorResponse.FromString,
        )
        self.Heartbeat = channel.stream_stream(
            "/indexify_coordinator.CoordinatorService/Heartbeat",
            request_serializer=coordinator__service__pb2.HeartbeatRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.HeartbeatResponse.FromString,
        )
        self.ListIndexes = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/ListIndexes",
            request_serializer=coordinator__service__pb2.ListIndexesRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.ListIndexesResponse.FromString,
        )
        self.GetIndex = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/GetIndex",
            request_serializer=coordinator__service__pb2.GetIndexRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.GetIndexResponse.FromString,
        )
        self.CreateIndex = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/CreateIndex",
            request_serializer=coordinator__service__pb2.CreateIndexRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.CreateIndexResponse.FromString,
        )
        self.GetExtractorCoordinates = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/GetExtractorCoordinates",
            request_serializer=coordinator__service__pb2.GetExtractorCoordinatesRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.GetExtractorCoordinatesResponse.FromString,
        )
        self.UpdateTask = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/UpdateTask",
            request_serializer=coordinator__service__pb2.UpdateTaskRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.UpdateTaskResponse.FromString,
        )
        self.ListStateChanges = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/ListStateChanges",
            request_serializer=coordinator__service__pb2.ListStateChangesRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.ListStateChangesResponse.FromString,
        )
        self.ListTasks = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/ListTasks",
            request_serializer=coordinator__service__pb2.ListTasksRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.ListTasksResponse.FromString,
        )
        self.GetSchema = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/GetSchema",
            request_serializer=coordinator__service__pb2.GetSchemaRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.GetSchemaResponse.FromString,
        )
        self.ListSchemas = channel.unary_unary(
            "/indexify_coordinator.CoordinatorService/ListSchemas",
            request_serializer=coordinator__service__pb2.GetAllSchemaRequest.SerializeToString,
            response_deserializer=coordinator__service__pb2.GetAllSchemaResponse.FromString,
        )


class CoordinatorServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def CreateContent(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetContentMetadata(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListContent(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CreateExtractionPolicy(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListExtractionPolicies(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CreateNS(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListNS(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetNS(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListExtractors(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def RegisterExecutor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Heartbeat(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListIndexes(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetIndex(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CreateIndex(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetExtractorCoordinates(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def UpdateTask(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListStateChanges(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListTasks(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetSchema(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListSchemas(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_CoordinatorServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "CreateContent": grpc.unary_unary_rpc_method_handler(
            servicer.CreateContent,
            request_deserializer=coordinator__service__pb2.CreateContentRequest.FromString,
            response_serializer=coordinator__service__pb2.CreateContentResponse.SerializeToString,
        ),
        "GetContentMetadata": grpc.unary_unary_rpc_method_handler(
            servicer.GetContentMetadata,
            request_deserializer=coordinator__service__pb2.GetContentMetadataRequest.FromString,
            response_serializer=coordinator__service__pb2.GetContentMetadataResponse.SerializeToString,
        ),
        "ListContent": grpc.unary_unary_rpc_method_handler(
            servicer.ListContent,
            request_deserializer=coordinator__service__pb2.ListContentRequest.FromString,
            response_serializer=coordinator__service__pb2.ListContentResponse.SerializeToString,
        ),
        "CreateExtractionPolicy": grpc.unary_unary_rpc_method_handler(
            servicer.CreateExtractionPolicy,
            request_deserializer=coordinator__service__pb2.ExtractionPolicyRequest.FromString,
            response_serializer=coordinator__service__pb2.ExtractionPolicyResponse.SerializeToString,
        ),
        "ListExtractionPolicies": grpc.unary_unary_rpc_method_handler(
            servicer.ListExtractionPolicies,
            request_deserializer=coordinator__service__pb2.ListExtractionPoliciesRequest.FromString,
            response_serializer=coordinator__service__pb2.ListExtractionPoliciesResponse.SerializeToString,
        ),
        "CreateNS": grpc.unary_unary_rpc_method_handler(
            servicer.CreateNS,
            request_deserializer=coordinator__service__pb2.CreateNamespaceRequest.FromString,
            response_serializer=coordinator__service__pb2.CreateNamespaceResponse.SerializeToString,
        ),
        "ListNS": grpc.unary_unary_rpc_method_handler(
            servicer.ListNS,
            request_deserializer=coordinator__service__pb2.ListNamespaceRequest.FromString,
            response_serializer=coordinator__service__pb2.ListNamespaceResponse.SerializeToString,
        ),
        "GetNS": grpc.unary_unary_rpc_method_handler(
            servicer.GetNS,
            request_deserializer=coordinator__service__pb2.GetNamespaceRequest.FromString,
            response_serializer=coordinator__service__pb2.GetNamespaceResponse.SerializeToString,
        ),
        "ListExtractors": grpc.unary_unary_rpc_method_handler(
            servicer.ListExtractors,
            request_deserializer=coordinator__service__pb2.ListExtractorsRequest.FromString,
            response_serializer=coordinator__service__pb2.ListExtractorsResponse.SerializeToString,
        ),
        "RegisterExecutor": grpc.unary_unary_rpc_method_handler(
            servicer.RegisterExecutor,
            request_deserializer=coordinator__service__pb2.RegisterExecutorRequest.FromString,
            response_serializer=coordinator__service__pb2.RegisterExecutorResponse.SerializeToString,
        ),
        "Heartbeat": grpc.stream_stream_rpc_method_handler(
            servicer.Heartbeat,
            request_deserializer=coordinator__service__pb2.HeartbeatRequest.FromString,
            response_serializer=coordinator__service__pb2.HeartbeatResponse.SerializeToString,
        ),
        "ListIndexes": grpc.unary_unary_rpc_method_handler(
            servicer.ListIndexes,
            request_deserializer=coordinator__service__pb2.ListIndexesRequest.FromString,
            response_serializer=coordinator__service__pb2.ListIndexesResponse.SerializeToString,
        ),
        "GetIndex": grpc.unary_unary_rpc_method_handler(
            servicer.GetIndex,
            request_deserializer=coordinator__service__pb2.GetIndexRequest.FromString,
            response_serializer=coordinator__service__pb2.GetIndexResponse.SerializeToString,
        ),
        "CreateIndex": grpc.unary_unary_rpc_method_handler(
            servicer.CreateIndex,
            request_deserializer=coordinator__service__pb2.CreateIndexRequest.FromString,
            response_serializer=coordinator__service__pb2.CreateIndexResponse.SerializeToString,
        ),
        "GetExtractorCoordinates": grpc.unary_unary_rpc_method_handler(
            servicer.GetExtractorCoordinates,
            request_deserializer=coordinator__service__pb2.GetExtractorCoordinatesRequest.FromString,
            response_serializer=coordinator__service__pb2.GetExtractorCoordinatesResponse.SerializeToString,
        ),
        "UpdateTask": grpc.unary_unary_rpc_method_handler(
            servicer.UpdateTask,
            request_deserializer=coordinator__service__pb2.UpdateTaskRequest.FromString,
            response_serializer=coordinator__service__pb2.UpdateTaskResponse.SerializeToString,
        ),
        "ListStateChanges": grpc.unary_unary_rpc_method_handler(
            servicer.ListStateChanges,
            request_deserializer=coordinator__service__pb2.ListStateChangesRequest.FromString,
            response_serializer=coordinator__service__pb2.ListStateChangesResponse.SerializeToString,
        ),
        "ListTasks": grpc.unary_unary_rpc_method_handler(
            servicer.ListTasks,
            request_deserializer=coordinator__service__pb2.ListTasksRequest.FromString,
            response_serializer=coordinator__service__pb2.ListTasksResponse.SerializeToString,
        ),
        "GetSchema": grpc.unary_unary_rpc_method_handler(
            servicer.GetSchema,
            request_deserializer=coordinator__service__pb2.GetSchemaRequest.FromString,
            response_serializer=coordinator__service__pb2.GetSchemaResponse.SerializeToString,
        ),
        "ListSchemas": grpc.unary_unary_rpc_method_handler(
            servicer.ListSchemas,
            request_deserializer=coordinator__service__pb2.GetAllSchemaRequest.FromString,
            response_serializer=coordinator__service__pb2.GetAllSchemaResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "indexify_coordinator.CoordinatorService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class CoordinatorService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def CreateContent(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/CreateContent",
            coordinator__service__pb2.CreateContentRequest.SerializeToString,
            coordinator__service__pb2.CreateContentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetContentMetadata(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/GetContentMetadata",
            coordinator__service__pb2.GetContentMetadataRequest.SerializeToString,
            coordinator__service__pb2.GetContentMetadataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListContent(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/ListContent",
            coordinator__service__pb2.ListContentRequest.SerializeToString,
            coordinator__service__pb2.ListContentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CreateExtractionPolicy(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/CreateExtractionPolicy",
            coordinator__service__pb2.ExtractionPolicyRequest.SerializeToString,
            coordinator__service__pb2.ExtractionPolicyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListExtractionPolicies(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/ListExtractionPolicies",
            coordinator__service__pb2.ListExtractionPoliciesRequest.SerializeToString,
            coordinator__service__pb2.ListExtractionPoliciesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CreateNS(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/CreateNS",
            coordinator__service__pb2.CreateNamespaceRequest.SerializeToString,
            coordinator__service__pb2.CreateNamespaceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListNS(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/ListNS",
            coordinator__service__pb2.ListNamespaceRequest.SerializeToString,
            coordinator__service__pb2.ListNamespaceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetNS(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/GetNS",
            coordinator__service__pb2.GetNamespaceRequest.SerializeToString,
            coordinator__service__pb2.GetNamespaceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListExtractors(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/ListExtractors",
            coordinator__service__pb2.ListExtractorsRequest.SerializeToString,
            coordinator__service__pb2.ListExtractorsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def RegisterExecutor(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/RegisterExecutor",
            coordinator__service__pb2.RegisterExecutorRequest.SerializeToString,
            coordinator__service__pb2.RegisterExecutorResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Heartbeat(
        request_iterator,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            "/indexify_coordinator.CoordinatorService/Heartbeat",
            coordinator__service__pb2.HeartbeatRequest.SerializeToString,
            coordinator__service__pb2.HeartbeatResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListIndexes(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/ListIndexes",
            coordinator__service__pb2.ListIndexesRequest.SerializeToString,
            coordinator__service__pb2.ListIndexesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetIndex(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/GetIndex",
            coordinator__service__pb2.GetIndexRequest.SerializeToString,
            coordinator__service__pb2.GetIndexResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CreateIndex(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/CreateIndex",
            coordinator__service__pb2.CreateIndexRequest.SerializeToString,
            coordinator__service__pb2.CreateIndexResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetExtractorCoordinates(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/GetExtractorCoordinates",
            coordinator__service__pb2.GetExtractorCoordinatesRequest.SerializeToString,
            coordinator__service__pb2.GetExtractorCoordinatesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def UpdateTask(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/UpdateTask",
            coordinator__service__pb2.UpdateTaskRequest.SerializeToString,
            coordinator__service__pb2.UpdateTaskResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListStateChanges(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/ListStateChanges",
            coordinator__service__pb2.ListStateChangesRequest.SerializeToString,
            coordinator__service__pb2.ListStateChangesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListTasks(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/ListTasks",
            coordinator__service__pb2.ListTasksRequest.SerializeToString,
            coordinator__service__pb2.ListTasksResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetSchema(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/GetSchema",
            coordinator__service__pb2.GetSchemaRequest.SerializeToString,
            coordinator__service__pb2.GetSchemaResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListSchemas(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/indexify_coordinator.CoordinatorService/ListSchemas",
            coordinator__service__pb2.GetAllSchemaRequest.SerializeToString,
            coordinator__service__pb2.GetAllSchemaResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )
